from enum import IntEnum


class NamedColor(IntEnum):
    aliceblue = 0
    antiquewhite = 1
    aqua = 2
    aquamarine = 3
    azure = 4
    beige = 5
    bisque = 6
    black = 7
    blanchedalmond = 8
    blue = 9
    blueviolet = 10
    brown = 11
    burlywood = 12
    cadetblue = 13
    chartreuse = 14
    chocolate = 15
    coral = 16
    cornflowerblue = 17
    cornsilk = 18
    crimson = 19
    cyan = 20
    darkblue = 21
    darkcyan = 22
    darkgoldenrod = 23
    darkgray = 24
    darkgreen = 25
    darkgrey = 26
    darkkhaki = 27
    darkmagenta = 28
    darkolivegreen = 29
    darkorange = 30
    darkorchid = 31
    darkred = 32
    darksalmon = 33
    darkseagreen = 34
    darkslateblue = 35
    darkslategray = 36
    darkslategrey = 37
    darkturquoise = 38
    darkviolet = 39
    deeppink = 40
    deepskyblue = 41
    dimgray = 42
    dimgrey = 43
    dodgerblue = 44
    firebrick = 45
    floralwhite = 46
    forestgreen = 47
    fuchsia = 48
    gainsboro = 49
    ghostwhite = 50
    gold = 51
    goldenrod = 52
    gray = 53
    grey = 54
    green = 55
    greenyellow = 56
    honeydew = 57
    hotpink = 58
    indianred = 59
    indigo = 60
    ivory = 61
    khaki = 62
    lavender = 63
    lavenderblush = 64
    lawngreen = 65
    lemonchiffon = 66
    lightblue = 67
    lightcoral = 68
    lightcyan = 69
    lightgoldenrodyellow = 70
    lightgray = 71
    lightgreen = 72
    lightgrey = 73
    lightpink = 74
    lightsalmon = 75
    lightseagreen = 76
    lightskyblue = 77
    lightslategray = 78
    lightslategrey = 79
    lightsteelblue = 80
    lightyellow = 81
    lime = 82
    limegreen = 83
    linen = 84
    magenta = 85
    maroon = 86
    mediumaquamarine = 87
    mediumblue = 88
    mediumorchid = 89
    mediumpurple = 90
    mediumseagreen = 91
    mediumslateblue = 92
    mediumspringgreen = 93
    mediumturquoise = 94
    mediumvioletred = 95
    midnightblue = 96
    mintcream = 97
    mistyrose = 98
    moccasin = 99
    navajowhite = 100
    navy = 101
    oldlace = 102
    olive = 103
    olivedrab = 104
    orange = 105
    orangered = 106
    orchid = 107
    palegoldenrod = 108
    palegreen = 109
    paleturquoise = 110
    palevioletred = 111
    papayawhip = 112
    peachpuff = 113
    peru = 114
    pink = 115
    plum = 116
    powderblue = 117
    purple = 118
    red = 119
    rosybrown = 120
    royalblue = 121
    saddlebrown = 122
    salmon = 123
    sandybrown = 124
    seagreen = 125
    seashell = 126
    sienna = 127
    silver = 128
    skyblue = 129
    slateblue = 130
    slategray = 131
    slategrey = 132
    snow = 133
    springgreen = 134
    steelblue = 135
    tan = 136
    teal = 137
    thistle = 138
    tomato = 139
    turquoise = 140
    violet = 141
    wheat = 142
    white = 143
    whitesmoke = 144
    yellow = 145
    yellowgreen = 146



import json
from bridges.symbol import *

class Rectangle(Symbol):
    """
    @brief This class defines a rectangle and is part of the symbol collection.
    
    A rectangle has height and width
    
    @author Matthew Mcquaigue
    @date 2018, 7/23/19
    
    \sa Shape collection tutorial, https://bridgesuncc.github.io/tutorials/Symbol_Collection.html
    """

    def __init__(self, **kwargs) -> None:
        """
        Constructor for the rectangle symbol
        Kwargs:
            (float) w: width of the rectangle
            (float) h: height of the rectangle
            (float) locx: x location of rectangle
            (float) locy: y loaction of rectangle
        Returns:
            None
        Raises:
            ValueError: if the width or height is less than 0
        """
        super(Rectangle, self).__init__()
        self._width = 1.0
        self._height = 1.0
        if 'w' in kwargs and 'h' in kwargs:
            if kwargs['w'] < 0 or kwargs['h'] < 0:
                raise ValueError("Illegal height or width! Height and Width values need to be positive")
            self._width = kwargs['w']
            self._height = kwargs['h']
        if 'locx' in kwargs and 'locy' in kwargs:
            self.locx = kwargs['locx']
            self.locy = kwargs['locy']
        else:
            self.locx = 0.0
            self.locy = 0.0


    def get_shape_type(self) -> str:
        """
        Getter for the name of the shape
        Returns:
             str: shape name
        """
        return "rect"

    @property
    def width(self) -> float:
        """
        Getter for the rectangle width
        Returns:
            float: the width
        """
        return self._width

    @width.setter
    def width(self, w) -> None:
        """
        Setter for the width of the rectangles
        Args:
            (float) w: the width to be applied to rectangle
        Returns:
            None
        Raises:
            ValueError: if the width is < 0 or > 300
        """
        if w <= 0:
            raise ValueError("Width needs to be in the range(0-300)")
        self._width = w

    @property
    def height(self) -> float:
        """
        Getter for the height of the rectangle
        Returns:
            float: the height of the rectangle
        """
        return self._height

    @height.setter
    def height(self, h) -> None:
        """
        Setter for the height of the rectangle
        Args:
            (float) h: the height to be applied
        Returns:
            None
        Raises:
            ValueError: if the height is < 0 or > 300
        """
        if h <= 0:
            raise ValueError("Height needs to be in the range(0-300)")
        self._height = h

    def set_rectangle(self, locx, locy, w, h):
        """
        Setter function for setting the rectangles size and location from scratch
        Args:
            (float) locx: the x location of rectangle
            (float) locy: the y location of rectangle
            (float) w: the width of rectangle
            (float) h: the height of rectangle
        Returns:
            None
        Raises:
            ValueError: if the height or width is < 0 or > 300
        """
        self.locx = locx
        self.locy = locy
        if w <= 0 or h <= 0:
            raise ValueError("Height, Width need to be in the range(0-300)")
        self.width = w
        self.height = h

    def get_json_representation(self) -> dict:
        """
        Getter function for the json representation of the data structure/shape
        Returns:
            dict: representing the json 
        """
        ds_json = super(Rectangle, self).get_json_representation()

        loc = [self.locx, self.locy]
        ds_json['lowerleftcorner'] = loc
        ds_json["width"] = self.width
        ds_json['height'] = self.height

        return ds_json







from bridges.bst_element import *


class AVLTreeElement(BSTElement):
    """
    @brief This class extends the BSTElement class by adding a height and balance factor fields that are useful in AVL trees.
    
    AVL tree elements include a 'height' and a 'balFactor' value,
    representing the height and balance factor of the AVL tree at
    that node, respectively. This is useful in representing
    AVL trees.
    
    AVLTree elements contain a visualizer (ElementVisualizer) object for setting visual
    attributes (color, shape, opacity, size), necessary for displaying them in a
    web browser.
    
    AVLTree elements also have a LinkVisualizer object, that is used when they are
    linked to another element, appropriate for setting link attributes, for instance,
    between the current element and its left or right child
    
    @author Kalpathi Subramanian, Mihai Mehedint, Matthew McQuaigue
    
    @date 6/22/16, 1/7/17, 5/17/17, 6/09/19
    
    Example tutorial using AVLTreeElement at: https://bridgesuncc.github.io/tutorials/AVL.html
    """
    def __init__(self, *args, **kwargs) -> None:
        """
        @brief AVL Tree constructor

        args:
            key, element
        kwargs:
            key: the search key
            e : the the specific object that the element is holding
        Returns:
            None
        """
        if len(args) != 0:
            if len(args) != 2:
                raise ValueError("Must contain 0 or 2 positional parameters")
            super(AVLTreeElement, self).__init__(key=args[0], e=args[1])

        if 'key' in kwargs and 'e' in kwargs:
            super(AVLTreeElement, self).__init__(key=kwargs['key'], e=kwargs['e'])
        elif 'key' in kwargs:
            super(AVLTreeElement, self).key = kwargs['key']
        elif 'e' in kwargs:
            super(AVLTreeElement, self).__init__(e=kwargs['e'])
        else:
            super(AVLTreeElement, self).__init__()
        self._height = self._bal_factor = 0

    def get_data_structure_type(self) -> str:
        """
        @brief Get the data structure type

        Returns:
             str: the type of data structure
        """
        return "AVLTree"

    @property
    def key(self) -> str:
        """
        @brief Getter for the search keys
        Returns:
            str: represeting the key
        """
        return super(AVLTreeElement, self).key

    @property
    def height(self) -> int:
        """
        @brief Getter for height of the avl tree at this node
        Returns:
            int: the height of the tree
        """
        return self._height

    @height.setter
    def height(self, value: int) -> None:
        """
        @brief Setter function for the height of the avl tree
        Args:
            (int) value: An integer for the height at this node
        Returns:
            None
        """
        self._height = value

    @property
    def balance_factor(self) -> int:
        """
        @brief Getter for the balance factor  of the tree at this node
        Returns:
            int: representing the balance factor
        """
        return self._bal_factor

    @balance_factor.setter
    def balance_factor(self, value: int) -> None:
        """
        @brief Setter for the balance factor of the tree at this node
        Args:
            (int) value: An integer for the balance factor at this node
        Returns:
            None
        """
        self._bal_factor = value

    @property
    def left(self):
        """
        @brief Getter for the left child of the avl tree
        Returns:
            child
        """
        return super(AVLTreeElement, self).left

    @left.setter
    def left(self, val) -> None:
        """
        @brief Setter for the left child of the avl tree
        Args:
            val: the value to be set for the left child to hold
        Returns:
            None
        """
        self.set_child(0, val)

    @property
    def right(self):
        """
        @brief Getter for the right child of the avl tree
        Returns:
            right child of this tree element
        """
        return super(AVLTreeElement, self).right

    @right.setter
    def right(self, val) -> None:
        """
        @brief Setter for the right child of the avl tree
        Args:
            val: the value to be set for the tight child to hold
        Returns:
            None
        """
        self.set_child(1, val)

    def get_element_representation(self) -> dict:
        """
        @brief Augment the element with the "height" and "balance factor" fields
        Returns:
            dict: representing the json of this element
        """
        orig_json_str = super(AVLTreeElement, self).get_element_representation()
        avl_json = {
            "height": self.height,
            "balance_factor": self.balance_factor
        }
        orig_json_str.update(avl_json)
        return orig_json_str




from bridges.symbol import *
from bridges.color import *

class SymbolGroup(Symbol):

    def __init__(self):
        super(SymbolGroup, self).__init__()
        self.al = []

    def add_symbol(self, s):
        self.al.append(s)

    def get_shape_type(self):
        return "group"

    def add_all_json(self, symbol_json, parent):
        id = len(symbol_json)

        obj = self.get_json_representation()

        obj['ID'] = id

        if parent is not None:
            obj['parentID'] = parent

        symbol_json.append(obj)

        for i in range(len(self.al)):
            self.al[i].add_all_json(symbol_json, id)


from bridges.symbol import *


class Text(Symbol):


    def __init__(self, label = None):
        super(Text, self).__init__()
        if label is not None:
            self._text = label
        else:
            self._text = ""

        self.stroke_width = 1.0

        self._font_size = None
        self._anchor_alignment_lr = None
        self._anchor_alignment_tb = None
        self._locx = 0.0
        self._locy = 0.0

    def get_shape_type(self):
        return "text"

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, t):
        self._text = t

    @property
    def font_size(self):
        return self._font_size

    @font_size.setter
    def font_size(self, s):
        if(s < 0.0):
            raise ValueError("Font size is too small")
        self._font_size = s

    def set_anchor_alignment(self, typeLR, typeTB):
        '''
        Sets the alignment parameters for the text label
        
        @param typeLR valid parameters are "left", "middle", and "right"
        @param typeTB valid parameters are "top", "bottom",  "embottom", "emtop", "middle"

        @return  symbol
        '''
        self._anchor_alignment_lr = typeLR
        self._anchor_alignment_tb = typeTB

    def set_anchor_location(self, x, y):
        self._locx = x
        self._locy = y

    def get_json_representation(self):

        json_builder = super(Text, self).get_json_representation()

        json_builder['anchor-location'] = [self._locx, self._locy]
        json_builder['text'] = self.text

        if self.font_size is not None:
            json_builder['font-size'] =self.font_size
        if self._anchor_alignment_lr is not None:
            json_builder['anchor-alignmentLR'] = self._anchor_alignment_lr
        if self._anchor_alignment_tb is not None:
            json_builder['anchor-alignmentTB'] = self._anchor_alignment_tb

        return json_builder







class Mesh:

    def __init__(self, vertices: list):
        self._vertices = vertices
        self._object_json = {
            'type': "custom mesh",
            'vertices': vertices
        }

    def push_representation(self, scene_json):
        scene_json['meshes'].append(self._object_json)


from bridges.element_visualizer import *
from bridges.link_visualizer import *
import re
from decimal import Decimal
import traceback
import sys


class Element:
    """
    @brief  This is the main superclass in BRIDGES for  deriving a number of
    elements used  in building data structures.
    
    This is the main superclass in BRIDGES for  deriving a number of
    elements used  in building data structures, viz., arrays, lists, trees and graphs.
    SLelement, DLelement, CircSLelement, CircDLelement, MLelement, TreeElement, BinTreeElement,
    BSTElement, AVLTreeElement and KDTreeElement are all subclasses
    (see class hierarchy above).  Element contains  two
    visualizer objects (ElementVisualizer, LinkVisualizer) for specifying
    visual attributes for nodes and links respectively. It also contains a label that
    that can be displayed in BRIDGES visualizations.
    
    All the tutorials under
    
    https://bridgesuncc.github.io/tutorials/Overview.html
    
    illustrate examples of using different types of Element objects and how to
    manipulate their visual attributes.
    
    @author Matthew Mcquaigue,  Kalpathi Subramanian
    
    2017, 2018, 6/24/19, 2020, 2021
    
    """
    ids = 0

    def get_data_structure_type(self) -> str:
        """
        Get the data structure representation
        Returns:
            str: representing the data structure type
        """
        return "Element"

    def __init__(self, **kwargs) -> None:
        """
        Element constructor. Creates an Element Visualizer and unique identifier for the current element.
        Args:
            (0) : "label" - the string label that is visible on the bridges Visualization
            (1) : "value" - data value (or object) E  used to construct Element
            (2) : "original" - element object to copy (if named "original")
            (3) : "color"  - color of element
            (4) : "opacity" - opacity of element
        Returns:
            None
        """
        Element.ids += 1
        self._link_visualizer = dict()
        self._ids = Element.ids
        self._identifier = str(self._ids)
        self._visualizer = ElementVisualizer()
        self._value = kwargs.get('val')
        self._label = kwargs.get('label') if kwargs.get('label') else "Default"

        if kwargs.get("color"):
            self.color = kwargs['color']

        if kwargs.get("opacity"):
            self.opacity = kwargs['opacity']

        if kwargs.get("original"):
            self._visualizer = ElementVisualizer(kwargs['original'].get_visualizer())
            self._label = str(kwargs['original'].label)
            self._value = kwargs['original']._value

    @property
    def value(self) -> object:
        """
        Getter for the value this element is holding
        Returns:
            generic object (application specific)
        """
        return self._value

    @value.setter
    def value(self, val) -> None:
        """
        Setter for the value of an element
        Args:
            val: the value (generic, application specific object)  this element will hold
        Returns:
            None
        """

        self._value = val

    @property
    def identifier(self) -> str:
        """
        Getter for the element identifier
        Return:
            str: element identifier (for internal use)
        """
        return self._identifier

    @identifier.setter
    def identifier(self, id: int) -> None:
        """
        Setter for the element identifier
        Args:
            id: the identifier (for internal use)
        Returns:
            None
        """
        self._identifier = id

    @property
    def visualizer(self) -> ElementVisualizer:
        """
        Getter for the element visualizer
        Returns:
            ElementVisualizer 
        """
        return self._visualizer

    @visualizer.setter
    def visualizer(self, vis: ElementVisualizer) -> None:
        """
        Setter function for this element visualizer
        Args:
            vis: the element visualizer
        Returns:
            None
        """
        self._visualizer = vis

    def get_link_visualizer(self, el) -> LinkVisualizer:
        """
        Getter for  the link visualizer object that links this element to 
        another element specified by the argument.
        Args:
            el: the element terminating the link
        Returns:
            LinkVisualizer: of this element
        Raises:
            ValueError: if an element is not passed in
        """
        if el in self._link_visualizer:
            return self._link_visualizer[el]
        else:
            self._link_visualizer[el] = LinkVisualizer()
            return self._link_visualizer[el]

    def set_link_visualizer(self, el) -> None:
        """
        Setter for the link visualizer of this element
        Args:
            (Element) el: the terminating element of this link;
             creates a new link visualizer for this link
        Returns:
            None
        """
        self._link_visualizer[el] = LinkVisualizer()

    def remove_link_visualizer(self, el) -> None:
        """
        Deleter function for the lik visualizer of this element
        Args:
            (Element) el: the terminating element of the link; the link visualizer 
            for this link will be deleted
        Returns:
            None
        """
        self._link_visualizer.pop(el)

    @property
    def label(self):
        """
        Getter for the element's label
        Returns:
           str: the element's label 
        """
        return self._label

    @label.setter
    def label(self, label):
        """
        Setter for the element's label
        Args:
           label: the element's label 
        Returns:
           None
        """
        self._label = label

    @property
    def size(self):
        """
        Getter for the element's size
        Returns:
           float: element size (0-50)
        """
        return self.visualizer.size

    @size.setter
    def size(self, sz):
        """
        Setter for the element's size
        Args:
           sz: the element's size (0.-50.)
        Returns:
           None
        """
        self.visualizer.size = sz

    @property
    def color(self):
        """
        Getter for the element's color
        Returns:
           Color: element's color
        """
        return self.visualizer.color

    @color.setter
    def color(self, col):
        """
        Setter for the element's size
        Args:
           col: the element's color
        Returns:
           None
        """
        self.visualizer.color = col

    @property
    def opacity(self):
        """
        Getter for the element's opacity
        Returns:
           float: element size (0-1.0)
        """
        return self.visualizer.opacity

    @opacity.setter
    def opacity(self, op):
        """
        Setter for the element's opacity
        Args:
           op: the element's size (0-1.0)
        Returns:
           None
        """
        self.visualizer.opacity = op

    @property
    def shape(self):
        """
        Getter for the element's shape type
        Returns:
           element shape
        """
        return self.visualizer.shape

    @shape.setter
    def shape(self, shp):
        """
        Setter for the element's shape
        Args:
           shp: the element's size (0-1.0)
        Returns:
           None
        """
        self.visualizer.shape = shp

    @property
    def id(self) -> int:
        """
        Get numer of ids of element object
        Returns:
           (int) number of ids of element
        """
        return self._ids

    def set_location(self, locX, locY):
        """
        Setter for the element's location
        Args:
           locX: the element's location in X
           locY: the element's location in Y
        Returns:
           None
        """
        self.visualizer.set_location(locX, locY)

    def get_locationX(self):
        """
        Getter for the element's location in X
        Returns:
           float: X coordinate of element's location
        """
        return self.visualizer.location_x

    def get_locationY(self):
        """
        Getter for the element's location in Y
        Returns:
           float: Y coordinate of element's location
        """
        return self.visualizer.location_y

    def get_element_representation(self):
        """
        Getter for the element's JSON representation (for internal use)
        Returns:
           str: JSON of element
        """
        json = {
            "name": str(self.label),
            "shape": str(self.visualizer.shape),
            "size": str(self.visualizer.size),
            "color": [str(self.visualizer.color.red), str(self.visualizer.color.green),
                      str(self.visualizer.color.blue), str(self.visualizer.color.alpha)]
        }
        loc_flag = not ((self.visualizer.location_x == Decimal('Infinity')) or (self.visualizer.location_y == Decimal('Infinity')))
        if loc_flag:
            json['location'] = [str(self.visualizer.location_x), str(self.visualizer.location_y)]
        return json

    def get_link_representation(self, lv, src, dest):
        """
        Getter for the JSON representation of  the element's link  (for internal use)
        Args:
           lv : link visualizer
           src: source vertex of link
           dest: destination vertex of link
        Returns:
           str: JSON of element's link from src to dest
        """
        link_json = lv.get_link_properties()
        link_json["source"] = src
        link_json["target"] = dest
        return link_json



class ActorMovieIMDB():
    """
    @brief A class to hold actor movie data -- using IMDB dataset.  

    This dataset has a set of actor-movie pairs with movie rating and
    genres, possibly. There are two datasets, the first has 1813 actor
    movie pairs and no other attribute data. The second also has
    movie ratings and genres.

    This object is generally not created by the user, to see how its created check 
    out bridges::data_src_dependent::data_source::get_actor_movie_imdb_data()

    @sa For an example, check out https://bridgesuncc.github.io/tutorials/Data_IMDB.html

    @author Kalpathi Subramanian, Matthew Mcquaigue

    @date 2/18/18, 12/29/20
    """
    
    def __init__(self, a = None, m = None, r = None, genr = None):
        """
        @brief This method initializes an actor movie object with an actor/movie pair.
        Args:
            a: actor  (string)
            m: movie  (string)
            r: rating  (float)
            genr: genres  (vector)
        """
        if a is not None:
            self._actor = a
        else:
            self._actor = ""
        if m is not None:
            self._movie = m
        else:
            self._movie = ""
        if r is not None:
            self._rating = r
        else:
            self._rating = float
        if genr is not None:
            self._genres = genr
        else:
            self._genres = []

    @property
    def actor(self):
        """
        @brief Get actor name

        Returns: 
            actor name (string)
        """
        return self._actor

    @actor.setter
    def actor(self, a):
        """
        Set actor name
        Args:
            a:  actor to set
        """
        self._actor = a

    @property
    def movie(self):
        """
        @brief Get movie name
        Returns:
            movie name 
        """
        return self._movie

    @movie.setter
    def movie(self, m):
        """
        @brief Set movie
        Args:
          m:  movie to set
        """
        self._movie = m

    @property
    def rating(self):
        """
        @brief Get movie rating
        Returns:
           Movie rating 
        """
        return self._rating

    @rating.setter
    def rating(self, r):
        """
        @brief Set movie rating
        Args:
            r:  rating to set
        """
        self._rating = r

    @property
    def genres(self):
        """
        @brief Get movie genres
        Returns:
            movie genres 
        """
        return self._genres

    @genres.setter
    def genres(self, g):
        """
        @brief Set movie genres
        Args:
            g:  genres to set (list of strings)
        """
        self._genres = g



#!/usr/bin/env python
from bridges.sl_element import *

class MLelement(SLelement):
    """
    @brief This class can be used to instantiate Multi-list Elements.
    
    This class extends SLelement (singly linked list element) to build multi-lists;
    Multilist elements contain a tag (boolean) that indicates if the element 
    contains a sublist or not; if the tag is true, then there is a sublist beginning
    at this node and the starting point is the `sublist' field in the element. 
    If the tag is false, then the list continues as a normal singly linked list. 
    The sublists are re recursive: any sublist can have its own sublists and so on.
    As in singly linked elements, the next pointer points to the following list element and 
    each element contains a generic application specific object.
    
    Multi-list elements contain a visualizer (ElementVisualizer) object for setting
    visual attributes (color, shape, opacity, size), necessary for displaying
    them in a web browser.
    
    Elements also have a LinkVisualizer object, that is used when they are linked to
    another element, appropriate for setting link attributes, for instance, between
    the current element and its next element. In this case, the link in question is that
    which connects the element to the following elements; a similar logic follows for
    sublists.
    
    @author Kalpathi Subramanian, Matthew McQuaigue
    
    @date 2018, 7/23/19, 2021
    
    \sa There is a tutorial about Multi Lists : https://bridgesuncc.github.io/tutorials/MultiList.html
    """

    def __init__(self, **kwargs) -> None:
        """
        Constructor for MLelement
        Args:
            (str) label: the label the SLelement will hold and show on bridges visualization
            (object) e: the generic object/value that this SLelement will hold
            (object) next: the next element that will be assigned to this SLelement next pointer
            sublist: the MLelement that is the beginning of a sublist
        Returns:
            None
        """
        if 'e' in kwargs:
            if 'label' in kwargs:
                super(MLelement, self).__init__(e=kwargs['e'], label=kwargs['label'])
            else:
                super(MLelement, self).__init__(e=kwargs['e'])
        else:
            super(MLelement, self).__init__()
        if 'next' in kwargs:
            super(MLelement, self).next = kwargs['next']
        if 'sublist' in kwargs:
            self._sublist = kwargs['sublist']
            self._tag = True
            self.set_link_visualizer(kwargs['sublist'])
        else:
            self._sublist = None
            self._tag = False

    @property
    def sub_list(self):
        """
        Getter for the sublist at this node if exists
        Returns:
            Element: the sublist head element
        """
        return self._sub_list

    @sub_list.setter
    def sub_list(self, sl) -> None:
        """
        Setter for the start of a new sublist
        Args:
            sl: the MLelement that is the beginning of a sublist
        Returns:
            None
        """
        self._sub_list = sl
        if sl is not None:
            self._tag = True
            self.set_link_visualizer(sl)
        self._default_sub_list_nodes()

    def _default_sub_list_nodes(self) -> None:
        """
        function for setting sublist defaults
        Returns:
            None
        """
        #by default, color and shape sublist nodes to distinguish them  from
        #remaining nodes
        self.visualizer.color = "red"
        self.visualizer.shape = "square"

    def get_data_structure_type(self) -> str:
        """
        Getter for the data structure type
        Returns:
            str: representing the type
        """
        return "MultiList"

    @property
    def next(self):
        """
        Retrieves the element following this element
        Returns:
            MLelement
        """
        return super(MLelement, self).next

    @next.setter
    def next(self, n):
        SLelement.next.fset(self, n)

    @property
    def tag(self) -> bool:
        """
        Getter for the tag of the element
        Returns:
             bool
        """
        return self._tag

    @tag.setter
    def tag(self, t) -> None:
        """
        Setter for the tag of the element
        Args:
            t: boolean value
        Returns:
            None
        """
        self._tag = t

    def get_data_structure_representation(self) -> dict:
        """
        Getter for the data structure representation
        Returns:
            dict: representing the json structure before dumping
        """
        nodes = []
        nodes.clear()
        self.get_list_elements(nodes)
        nodes_JSON = []
        node_map = dict()
        for i in range(0, len(nodes)):
            node_map[nodes[i]] = i
            nodes_JSON.append(nodes[i].get_element_representation())
        links_JSON = []
        for j in range(0, len(nodes)):
            par = nodes[j]
            if par.tag:
                chld = par.sub_list
                #  sub list
                if chld is not None:
                    #  add the link
                    links_JSON.append(self.get_link_representation(par.get_link_visualizer(chld),
                                                                   str(node_map.get(par)), str(node_map.get(chld))))
            chld = par.next
            if chld is not None:
                #  add the link
                links_JSON.append(self.get_link_representation(par.get_link_visualizer(chld),
                                                               str(node_map.get(par)), str(node_map.get(chld))))
        json_dict = {
            "nodes": nodes_JSON,
            "links": links_JSON
        }
        return json_dict

    def get_list_elements(self, nodes):
        """
        Getter for the elements of the list
        Args:
            nodes: a list of the nodes
        Returns:
            element
        """
        self._get_list_elements_R(self, nodes)


    def _get_list_elements_R(self, list, nodes):
        el = list
        while el is not None:
            nodes.append(el)
            if el.tag:
                self._get_list_elements_R(el.sub_list, nodes)
            el = el.next



#!/usr/bin/env python
import base64
import wave
import json
import array
import math
from bridges.audio_channel import AudioChannel

class AudioClip(object):
    """
    @brief This class provides support for reading, modifying, and playing, audio waveform.
    
    This class provides a way to represent an AudioClip (think of a
    WAV file) in Bridges as waveforms.
    
    An AudioClip can be composed of multiple channels: a stereo sound
    would be composed of 2 channels (Left and Right), a mono sound
    would be composed of a single channel. A 5.1 sound would be
    composed of 6 channels. When building an AudioClip from a file, the
    number of channels is taken from the file; some constructors have a
    num_channels parameter that enables to pass the number of channels
    explicitly. If unsure, one can know how many channels are in an
    audio clip using get_num_channels().
    
    Each channel is essentially a 1D signal. That is to say, it is an
    array of values that represent how far the membrane of a speaker
    should be from its resting position. The quality of the sound is
    controlled by two parameters: sampling rate and sampling depth.
    
    Sampling rate tells how many positions per second are encoded by
    the AudioClip. It is expressed in Hertz. CD quality is 44100Hz;
    while walkie-talkies use 8000Hz. It is set automatically if read
    from a file; or it can be passed as the sampleRate parameter to
    some of the constructors. The sampling rate can be obtained from an
    AudioClip using get_sample_rate().
     
    The length of an AudioClip is expressed in number of samples. So if
    an AudioClip is composed of 16,000 samples with a sampling rate of
    8000Hz, the clip would be 2 seconds long. The number of samples
    can obtained with get_sample_count(); it is set from a file or can be
    passed as the sampleCount parameter of some of the constructor.
    
    The sampling depth indicates how many different positions the
    membrane can take. It is typically expressed in bits with supported
    values being 8-bit, 16-bit, 24-bit, and 32-bit. If a clip is
    encoded with a depth of 8 bits, the membrane can take 2^8 different
    position ranging from -128 to +127, with 0 being the resting
    position. The sampling depth is read from files or passed as the
    sampleBits parameter of the constructor. The sampling depth of an
    existing clip can be obtained with get_sample_bits().
    
    The individual samples are accessed with the get_sample() and
    set_sample() functions. The samples are integer values in the
    2^(get_sample_bits()-1) ; 2^(get_sample_bits()-1)[ range. The
    functions allow to specify for channel and sample index.
    
    @author Luke Sloop, Erik Saule
    
    @date 2020, 1/31/2020, 2021
    """
    def __init__(self, filepath: str="", sample_count: int=0, num_channels: int=1, sample_bits: int=32, sample_rate: int=44100) -> None:
        """
        AudioBase constructor.
        specify either a filepath or all the other parameters.
        Args:
            (str) filepath: name of the wav file to creat a clip of. If this parameter is used, all the other ones are ignored.
            (int) sample_count: The total number of samples in this audio object
            (int) num_channels: number of channels (stereo would be 2)
            (int) sample_rate: The number of samples in 1 second of audio (default to cd quality: 44100)
            (int) sample_bits: Bit depth, that is to say, each sample will be in the [-2^(get_sample_bits()-1) ; 2^(get_sample_bits()-1)[ range
        Returns:
            None
        """
        if filepath != "":
            self._from_filepath(filepath)
            return

        if sample_count > 1000000000:
            raise ValueError("Maximum frames exceeded with value %d" % self.get_sample_count())

        if sample_bits != 8 and sample_bits != 16 and sample_bits != 24 and sample_bits != 32:
            raise ValueError("sample_bits should be 8, 16, 24, or 32")

        if num_channels <= 0:
            raise ValueError("num_channels should be positive")

        if sample_rate <= 0:
            raise ValueError("sample_rate should be positive")
        
        self.sample_count = sample_count
        self.sample_rate = sample_rate
        self.sample_bits = sample_bits

        self.num_channels = num_channels
        # Initialize the channels
        self._channels = []
        for i in range(self.num_channels):
            self._channels.append(AudioChannel(sample_count=self.sample_count, sample_bits=self.sample_bits)) 

    def _from_filepath(self, filepath: str):
        with wave.open(filepath, "r") as f:
            self.__init__(sample_count=f.getnframes(), sample_rate=f.getframerate(), num_channels=f.getnchannels(), sample_bits=f.getsampwidth()*8)

            self.framebytes = f.readframes(f.getnframes())
            framebytes = self.framebytes

            channel = 0
            count = 0
            for i in range(0, len(framebytes), self.get_sample_bytes()):
                if self.get_sample_bytes() == 1:
                    val = int.from_bytes(framebytes[i:i+self.get_sample_bytes()], byteorder='little', signed=False)
                    val = val - 128
                    self.set_sample(channel, count, val)
                else:
                    val = int.from_bytes(framebytes[i:i+self.get_sample_bytes()], byteorder='little', signed=True)
                    self.set_sample(channel, count, val)
                
                channel += 1

                if channel >= f.getnchannels():
                    count += 1
                    channel = 0

    def get_num_channels(self) -> int:
        """
        Return the number of channels in this AudioClip. 1 for mono, 2 for stereo, etc.
        Returns:
            int: The number of channels of audio samples this object holds.
        """
        return self.num_channels

    def get_channel(self, index: int) -> AudioChannel:
        """
        Get the audio channel at index. The index should be less than get_num_channels().
        Args:
            (int) index: The index of the channel to get. 0 for front-left, 1 for front-right, etc.
        Returns:
            AudioChannel: The audio channel at index
        """
        return self._channels[index]

    def get_sample_rate(self) -> int:
        """
        Get the sample rate of this audio clip. This is the number of samples that are taken in one second.
        Returns:
            int: The sample rate or number of samples in 1 second of audio
        """
        return self.sample_rate

    def get_sample_count(self) -> int:
        """
        Get the number of samples in each channel of this audio object. 
        Each channel will contain this number of samples.
        Returns:
            int: The total number of samples in this audio object
        """
        return self.sample_count

    def get_sample(self, channel: int, index: int) -> int:
        """
        Get the sample at the index of the sample data from a specific channel.
        Args:
            (int) channel: The index of the channel to get. 0 for front-left, 1 for front-right, etc.
            (int) index: The index of the sample to get. From 0 - get_sample_count()
        Returns:
            int: The sample in the [-2^(get_sample_bits()-1) ;  2^(get_sample_bits()-1)) range
        """
        value = self.get_channel(channel).get_sample(index)

        return int(value)

    def set_sample(self, channel: int, index: int, value: int) -> None:
        """
        Set the sample at the index of the sample data to value
        Args:
            (int) channel: The index of the channel to get. 0 for front-left, 1 for front-right, etc.
            (int) index: The index of sampledata to set which must be less than get_sample_count()
            (int) value: The value to set the sample to which must be a valid signed integer with bit length get_sample_bits(). That is to say in the [-2^(get_sample_bits()-1) ;  2^(get_sample_bits()-1)) range).
        Returns:
            None
        """

        if (value < -2**(self.get_sample_bits()-1)) or (value >= 2**(self.get_sample_bits()-1)):
            raise ValueError("Audio value Out of Bound. Should be in [-2^(get_sample_bits()-1) ;  2^(get_sample_bits()-1)) range")
        
        self.get_channel(channel).set_sample(index, int(value))

    def get_sample_bits(self) -> int:
        """
        Get the number of bits for the samples in this audio clip.  Will be 8, 16, 24, or 32 bits.
        Returns:
            int: The number of bits for each sample 
        """
        return self.sample_bits

    def get_sample_bytes(self) -> int:
        """
        Get the number of bytes for the samples in this audio clip.  Will be 1, 2, 3, or 4 bits.
        Returns:
            int: The number of bytes for each sample 
        """
        return self.sample_bits // 8

    def _get_type_code(self) -> str:
        if self.get_sample_bytes() == 1:
            return "b"
        elif self.get_sample_bytes() == 2:
            return "h"
        elif self.get_sample_bytes() == 3:
            return "f"
        elif self.get_sample_bytes() == 4:
            return "l"
        else:
            raise ValueError("Wave file sample bytes of unsupported length %d, supported lengths are 8, 16, 24, and 32 bit" % (self.get_sample_bytes() * 8))

    def get_data_structure_type(self) -> str:
        """
        Get the data structure type
        Returns:
            str : data structure type
        """
        return "Audio"

    def get_data_structure_representation(self) -> dict:
        """ Return a dictionary of the data in this audio file
        Returns:
            dict: The data of this audio file
        """
        json_dict = {}

        json_dict["encoding"] = "RAW"
        json_dict["numChannels"] = self.num_channels
        json_dict["sampleRate"] = self.get_sample_rate()
        json_dict["bitsPerSample"] = self.get_sample_bits()
        json_dict["numSamples"] = self.get_sample_count()

        # Combine all channel data
        framedata = []
        for i in range(self.sample_count):
            for c in range(self.num_channels):
                # Go straight to channel sample for correct bit data
                framedata.append(self._channels[c].get_sample(i))

        if self.get_sample_bytes() == 4:
            newarr = []
            for val in framedata:
                minmax32 = (2 ** 32 / 2.0) - 1
                minmax16 = (2 ** 16 / 2.0) - 1

                newval = (val / minmax32) * minmax16

                newarr.append(int(newval))

            json_dict["bitsPerSample"] = 16
            json_dict["samples"] = base64.b64encode(array.array("h", newarr).tobytes()).decode("utf-8")
        elif self.get_sample_bytes() != 3:
            json_dict["samples"] = base64.b64encode(array.array(self._get_type_code(), framedata).tobytes()).decode("utf-8")
        else:
            shiftedbytes = bytearray()
            for sample in framedata:
                shiftedbytes += int.to_bytes(sample, length=3, byteorder='little', signed=True)

            json_dict["samples"] = base64.b64encode(shiftedbytes).decode("utf-8")

        return json_dict

    def display(self) -> None:
        """ Print information about this audio file to the console
        """
        print("Num Channels: %d, Sample Rate: %d, Sample Bits: %d, Num Samples: %d" % (self.num_channels, self.sample_rate, self.get_sample_bits(), self.get_sample_count()))

    def audio_from_json(json_dict: dict) -> 'AudioClip':
        """ Create an AudioClip from a json dictionary created by another AudioClip object
        Args:
            (dict) json_dict: The json dictionary created by another AudioClip object
        """
        audio = AudioClip(sample_count=json_dict["numSamples"], num_channels=json_dict["numChannels"], sample_rate=json_dict["sampleRate"], sample_bits=json_dict["bitsPerSample"])
        
        data = []
        if audio.get_sample_bytes() != 3:
            data = array.array(audio._get_type_code(), base64.b64decode(json_dict["samples"]))
        else:
            # No simple type to convert 24 bit base64 string to integer array
            chunk = base64.b64decode(json_dict["samples"])

            for i in range(0, len(chunk), audio.get_sample_bytes()):
                data.append(int.from_bytes(chunk[i:i+audio.get_sample_bytes()], byteorder='little', signed=True))

        for i in range(audio.get_sample_count() * audio.get_num_channels()):
            audio.set_sample(i % audio.get_num_channels(), int(i / audio.get_num_channels()), (data[i] / ((2 ** audio.get_sample_bits() / 2) - 1) * ((2 ** 32 / 2) - 1)))

        return audio



#!/usr/bin/env python
from bridges.sl_element import *

class DLelement(SLelement):
    """
    @brief This class is used to create doubly linked element objects.
    
    This class extends Element and takes a generic parameter <E> representing
    application specific data. This element forms the basic building block for
    doubly linked lists. Doubly linked elements have two links,
    "next" and "previous", that point to the previous and succeeding nodes along the list.
    
    Elements contain a visualizer (ElementVisualizer) object for setting visual
    attributes (color, shape, opacity, size), necessary for displaying them in a web
    browser.
    
    Elements also have a LinkVisualizer object that is used when they are linked to
    another element, appropriate for setting link attributes, such as in linked lists,
    between the current element and its next or previous nodes.
    
    @author Mihai Mehedint, Kalpathi Subramanian
    @date  2018, 6/24/19
    
    \sa Doubly Linked List tutorial : https://bridgesuncc.github.io/tutorials/DoublyLinkedList.html
    """


    def __init__(self, *args, **kwargs) -> None:
        """
        Constructor for DLelement
        Kwargs:
            (object) e: the genereic application specific object that DLelement is holding
            next: the DLelement that should be assigned to the next pointer
            prev: the DLelement that should be assigned to the prev pointer
            (str) label: the label for this DLelement
        Return:
            None
        """
        super(DLelement, self).__init__(**kwargs)
        self._next = kwargs.get('next')
        self._prev = kwargs.get('prev')


    def get_data_structure_type(self) -> str:
        """
        This method gets the data structure type
        Returns:
            str: representing the data structure type
        """
        return "DoublyLinkedList"

    @property
    def next(self):
        """
        Getter for the next element in Dlelement
        Returns:
            element pointed to by this element
        """
        return self._next

    @next.setter
    def next(self, el):
        """
        Setter for the next element
        Args:
            el: element to be set 
        Returns:
            None
        """
        self._next = el

    @property
    def prev(self):
        """
        Getter for the prev element in Dlelement
        Returns:
            element that is prior to this element
        """
        return self._prev

    @prev.setter
    def prev(self, el):
        """
        Setter for the prev element
        Args:
            el: element to be set

        Returns:
            None
        """
        #remove any existing link visualizer from this node
        if self._prev is not None:
            self.remove_link_visualizer(self._prev)
        self._prev = el
        #  add a new link visualizer
        if el is not None:
            self.set_link_visualizer(el)

    def get_data_structure_representation(self) -> dict:
        """
        Getter for the json representation of this list
        Returns:
            dict: representing the JSON format
        """
        #  map to reorder the nodes for building JSON
        node_map = dict()
        #  get teh list nodes
        nodes = []
        SLelement.get_list_elements(self, nodes)
        #  generate the JSON of the list nodes
        nodes_JSON = []
        k = 0
        while k < len(nodes):
            node_map[nodes[k]] = k
            nodes_JSON.append(nodes[k].get_element_representation())
            k += 1
        links_JSON = []
        k = 0
        while k < len(nodes):
            par = nodes[k]
            nxt = par.next
            prv = par.prev
            if nxt is not None:
                #  add the link
                links_JSON.append(self.get_link_representation(par.get_link_visualizer(nxt), str(node_map.get(par)), str(node_map.get(nxt))))
            if prv is not None:
                #  add the link
                links_JSON.append(self.get_link_representation(par.get_link_visualizer(prv), str(node_map.get(par)), str(node_map.get(prv))))
            k += 1
        #  add the link
        json_dict = {
            "nodes": nodes_JSON,
            "links": links_JSON
        }
        return json_dict

    def reverse_iterator(self):
        """
        This iterator is used with range loops
        """
        return DLelementReverseIterator(self)

class DLelementReverseIterator():

    def __init__(self, current):
        self.current = current

    def has_next(self):
        return self.current is not None

    def next(self):
        ret = self.current.value
        self.current = self.current.prev
        return ret



from bridges.grid import *
from bridges.game_cell import *
import base64


class GameGrid(Grid):
    """
    This is a class in BRIDGES for representing an (m x n) grid. Each position in the grid will hold a GameCell object, each of which has a foreground color, background color, and a symbol.
    @author David Burlinson, Matthew McQuaigue
    """

    def set_encoding(self, encoding: str) -> None:
        """
        Enable changing the game grid encoding when building JSON representation.
        Args:
            (str) encoding: type of encoding. Supports "raw" and "rle"
        """
        if encoding == "raw" or encoding == "rle":
            self.encoding = encoding
        else:
            print("Unrecognized encoding \'" + encoding +
                  "\', defaulting to raw. Options: raw, rle")
            self.encoding = "raw"

    def get_data_structure_type(self):
        return "GameGrid"

    def __init__(self, rows = None, cols = None):
        """
        Grid Constructor
        Args:
            rows: representing the number of rows of the grid
            cols: representing the number of columns of the grid
        """
        self.bf_bg = bytearray()
        self.bf_fg = bytearray()
        self.bf_symbols = bytearray()
        self.encoding = "raw"
        if rows is None and cols is None:
            super(GameGrid, self).__init__(rows=30, cols=30)
            self.grid_size = [30, 30]
        else:
            super(GameGrid, self).__init__(rows=rows, cols=cols)
            self.grid_size = [rows, cols]
        self.initialize_game_Grid()

    def initialize_game_Grid(self) -> None:
        """
        Populate the grid with default game cells
        """
        for i in range(self.grid_size[0]):
            for j in range(self.grid_size[1]):
                self.set(i, j, GameCell())
        self.bf_bg = bytearray(self.grid_size[0]*self.grid_size[1])
        self.bf_fg = bytearray(self.grid_size[0]*self.grid_size[1])
        self.bf_symbols = bytearray(self.grid_size[0]*self.grid_size[1])

    def get_bg_color(self, row: int, col: int):
        """
        Get the background color at cell row, col
        Args:
            row: row index to get color
            col: col index to get color
        Returns:
            Color at row, col
        """
        return super(GameGrid, self).get(row, col).bg

    def set_bg_color(self, row: int, col: int, color) -> None:
        """
        Set background color of a cell using an enum argument
        Args:
            row: row index to set color
            col: col index to set color
            color: Named Color enum argument to set the background at the chosen position
        """
        if type(color) == NamedColor:
            self.get(row, col).bg = color
        else:
            self.get(row, col).bg = NamedColor[color]

    def get_fg_color(self, row: int, col: int):
        """
        Get the foreground color at cell row, col
        Args:
            (int) row: row index to get color
            (int) col: col index to get color
        Returns:
            Color at row, col
        """
        return self.get(row, col).fg

    def set_fg_color(self, row: int, col: int, color):
        """
        Set foreground color of a cell using an enum argument
        Args:
            row: row index to set color
            col: col index to set color
            color: Named Color enum argument to set the background at the chosen position
        """
        if type(color) == NamedColor:
            self.get(row, col).fg = color
        else:
            self.get(row, col).fg = NamedColor[color]

    def get_symbol(self, row: int, col: int):
        """
        Get the symbol at cell row, col
        Args:
            row: row index to get color
            col: col index to get color
        """
        return self.get(row, col).symbol

    def get_symbol_color(self, row, col):
        """
        Get the symbol color at row,col
        Args:
            (int) row: row index to get color
            (int) col: col index to get color
        """
        return self.get(row, col).fg

    def draw_symbol(self, row, col, symbol, color):
        """
        Draw a symbol at the location of row,col with color
        Args:
            row: row index to set color
            col: col index to set color
            symbol: symbol argument to set the symbol at the chosen position
            color: Named Color enum argument to set the background at the chosen position
        """
        self.get(row, col).symbol = symbol
        self.get(row, col).fg = color

    def get_data_structure_representation(self) -> dict:
        """
        Get the JSON representation of the game grid. Contains separate foreground, background, and symbol arrays
        Returns:
            dict: represnting the game_grids json
        """
        count = 0

        json_dict = {
            "encoding": self.encoding,
            "dimensions": [self.grid_size[0], self.grid_size[1]]
        }

        if self.encoding == "rle":
            bg = []
            fg = []
            symbols = []

            for i in range(self.grid_size[0]):
                if self.grid[i] is not None:
                    for j in range(self.grid_size[1]):
                        if self.grid[i][j] is not None:
                            gc = self.grid[i][j]
                            bg.append(gc.bg.value)
                            fg.append(gc.fg.value)
                            symbols.append(gc.symbol.value)
                            count += 1
            json_dict['bg'] = self.run_length(bg)
            json_dict['fg'] = self.run_length(fg)
            json_dict['symbols'] = self.run_length(symbols)

        if self.encoding == "raw":
            for i in range(self.grid_size[0]):
                if self.grid[i] is not None:
                    for j in range(self.grid_size[1]):
                        if self.grid[i][j] is None:
                            gc = self.grid[i][j]
                            self.bf_bg.append(gc.get_bg_byte())
                            self.bf_fg.append(gc.get_fg_byte())
                            self.bf_symbols.append(gc.get_symbol_byte())

            json_dict['bg'] = base64.b64encode(self.bf_bg)
            json_dict['fg'] = base64.b64encode(self.bf_fg)
            json_dict['symbols'] = base64.b64encode(self.bf_symbols)
        return json_dict

    def run_length(self, arr):
        count = 1
        out = str()
        for i in range(1, len(arr)):
            if arr[i-1] == arr[i]:
                count += 1
                if len(arr) - i == 1:
                    out += str(arr[i]) + "x" + str(count)
            else:
                out += str(arr[i-1]) + "x" + str(count) + ","
                count = 1
                if len(arr) - i == 1:
                    out += str(arr[i]) + 'x' + str(count)

        return out





#!/usr/bin/env python
from typing import List

import webcolors


class Color(object):
    """
    This class is used to represent colors in bridges.

    We use and RGBA model to represent colors, with the Red Green and Blue components ranging from 0-255,
    with the alpha ranging from 0.0-1.0 inclusive.

    We use webcolors to handle color names passed to the constructor/set_color function.
    https://www.w3.org/TR/css-color-3/#svg-color

    Color names are listed below:<br>

    aliceblue, antiquewhite, cyan, aquamarine, azure, beige, bisque, 
    black, blanchedalmond, blue, blueviolet, brown, burlywood, cadetblue, 
    chartreuse, chocolate, coral, cornflowerblue, cornsilk, crimson, 
    darkblue, darkcyan, darkgoldenrod, darkgrey, darkgreen, darkkhaki, 
    darkmagenta, darkolivegreen, darkorange, darkorchid, darkred, 
    darksalmon, darkseagreen, darkslateblue, darkslategrey, darkturquoise, 
    darkviolet, deeppink, deepskyblue, dimgrey, dodgerblue, firebrick, 
    floralwhite, forestgreen, magenta, gainsboro, ghostwhite, gold, 
    goldenrod, grey, green, greenyellow, honeydew, hotpink, indianred, 
    indigo, ivory, khaki, lavender, lavenderblush, lawngreen, 
    lemonchiffon, lightblue, lightcoral, lightcyan, lightgoldenrodyellow, 
    lightgrey, lightgreen, lightpink, lightsalmon, lightseagreen, 
    lightskyblue, lightslategrey, lightsteelblue, lightyellow, lime, 
    limegreen, linen, maroon, mediumaquamarine, mediumblue, mediumorchid, 
    mediumpurple, mediumseagreen, mediumslateblue, mediumspringgreen, 
    mediumturquoise, mediumvioletred, midnightblue, mintcream, mistyrose, 
    moccasin, navajowhite, navy, oldlace, olive, olivedrab, orange, 
    orangered, orchid, palegoldenrod, palegreen, paleturquoise, 
    palevioletred, papayawhip, peachpuff, peru, pink, plum, 
    powderblue, purple, red, rosybrown, royalblue, saddlebrown, 
    salmon, sandybrown, seagreen, seashell, sienna, silver, 
    skyblue, slateblue, slategrey, snow, springgreen, steelblue, tan, 
    teal, thistle, tomato, turquoise, violet, wheat, white, whitesmoke, 
    yellow, yellowgreen

    Attributes: red (int): red component of color ranging from 0-255 inclusive (default 0) green (int): green component of color ranging from 0-255 inclusive (default 0) blue (int): blue component of color ranging from 0-255 inclusive (default 0) alpha (float): alpha component of color ranging from 0.0-1.0 inclusive (default 1.0) rgba (tuple(int, int, int, alpha)): RGBA components as respective tuple
    Args:
        args: int, int, int, Optional(float) or a str as singular arg
        kwargs:
            r or red: Optional(int)
            b or blue: Optional(int)
            g or green: Optional(int)
            a or alpha: Optional(float)
            col_name: Optional(str)
    Raises:
        ValueError: if args is not 3 ints with an optional 4th arg for alpha or just one str arg
        ValueError: if a str passed is not a valid webcolor
        ValueError: if any of the RGBA values are outside of their respective range
    Examples:
        >>> my_color = Color("red")
        >>> my_color.rgba
        (255, 0, 0, 1.0)
        >>> my_color = Color(r=255)
        >>> my_color.rgba
        (255, 0, 0, 1.0)
        >>> my_color = Color(255, 0, 0)
        >>> my_color.rgba
        (255, 0, 0, 1.0)
        >>> my_color = Color()
        >>> my_color.red = 255
        >>> my_color.rgba
        (255, 0, 0, 1.0)
    """
    @property
    def red(self) -> int:
        """
        Getter for red component of color (0-255 inclusive)
        Returns:
            (int) red component of color
        """
        return self._red

    @red.setter
    def red(self, value: int):
        """
        Setter for red component of color (0-255 inclusive)
        Args:
            value(int): red component of color
        Returns:
            None
        """
        try:
            value = int(value)
        except ValueError:
            raise ValueError("Value for RGB attributes must be able to be casted to int")
        if value < 0 or value > 255:
            raise ValueError("Value for RGB attributes should range from 0-255 inclusive")

        self._red = value

    @property
    def green(self) -> int:
        """
        Getter for green component of color (0-255 inclusive)
        Returns:
            int: green component of color
        """
        return self._green

    @green.setter
    def green(self, value: int): 
        """
        Setter for green component of color (0-255 inclusive)
        Args:
            value(int): green component of color
        Returns:
            None
        """
        try:
            value = int(value)
        except ValueError:
            raise ValueError("Value for RGB attributes must able be to be casted to int")
        if value < 0 or value > 255:
            raise ValueError("Value for RGB attributes should range from 0-255 inclusive")

        self._green = int(value)

    @property
    def blue(self) -> int:
        """
        Getter for blue component of color (0-255 inclusive)
        Returns:
            int: blue component of color
        """
        return self._blue

    @blue.setter
    def blue(self, value: int):
        """
        Setter for blue component of color (0-255 inclusive)
        Args:
            value(int): blue component of color
        Returns:
            None
        """
        try:
            value = int(value)
        except ValueError:
            raise ValueError("Value for RGB attributes must able be to be casted to int")
        if value < 0 or value > 255:
            raise ValueError("Value for RGB attributes should range from 0-255 inclusive")

        self._blue = int(value)

    @property
    def alpha(self) -> float:
        """
        Getter for alpha(opacity) component of color (0-1.0 inclusive)
        Returns:
            float: alpha component of color
        """
        return self._alpha

    @alpha.setter
    def alpha(self, value: float):
        """
        Setter for alpha component of color (0-1.0 inclusive)
        Args:
            value(float): alpha component of color (0-1.0)
        Returns:
            None
        """
        try:
            value = float(value)
        except ValueError:
            raise ValueError("Value for Alpha must be able to be casted to float")
        if value < 0.0 or value > 1.0:
            raise ValueError("Value for alpha should range from 0.0 - 1.0 inclusive")

        self._alpha = value

    @property
    def rgba(self) -> (int, int, int, float):
        """
        RGBA components as respective tuple.  Represents the RGBA values of the color as a tuple, can be used to set or get all values at once
        Returns:
            (int, int, int, float) - RGBA values respectively
        """
        return self.red, self.green, self.blue, self.alpha

    @rgba.setter
    def rgba(self, rgba: (int, int, int, float)):
        """
        Set RGBA components.
        Args:
            rgba: r,g,b,a (list)
		Returns: 
            None
        """
        self.red, self.blue, self.green, self.alpha = rgba

    def __init__(self, *args, **kwargs):
        """ Constructor for a Color object
        Usage: requires either 3 ints 0-255 for RGB and an optional float 0.0-1.0 for alpha or a str of a web color can also key the RGBA values with r, g, b, a or red, green, blue, alpha respectively and col_name for the str
        Args:
            args: int, int, int, optional float or just a str
            kwargs: r/red: int, b/blue: int, g/green: int optional a/alpha: float or col_name: str
        Returns:
            None
        """
        self._red = 0
        self._green = 0
        self._blue = 0
        self._alpha = 1.0
        self.set_color(*args, **kwargs)

    def set_color(self, *args, **kwargs) -> None:
        """
        Sets color for a an element or link. Requires either 3 ints 0-255 for RGB and an optional float 0.0-1.0 for alpha or a str of a web color can also key the RGBA values with r, g, b, a or red, green, blue, alpha respectively and col_name for the str
        Args:
            args: int, int, int optional float or str
            kwargs: r/red: int, b/blue: int, g/green: int optional a/alpha: float or col_name: str
        """
        col_name = None
        if args:
            errorcondition = True
            if len(args) == 1 and type(args[0]) == list:
                self.red = args[0][0]
                self.green = args[0][1]
                self.blue = args[0][2]
                self.alpha = args[0][3]
                errorcondition=False
            if len(args) == 4 or len(args) == 3:
                self.red = args[0]
                self.green = args[1]
                self.blue = args[2]
                if len(args) == 4:
                    self.alpha = args[3]
                errorcondition = False
            elif len(args) == 1:
                if type(args[0]) is str:
                    col_name = args[0]
                    errorcondition=False
                elif type(args[0]) is Color:
                    self.red = args[0].red
                    self.green = args[0].green
                    self.blue = args[0].blue
                    self.alpha = args[0].alpha
                    errorcondition = False
            if errorcondition:
                raise ValueError("To use Color constructor pass 3 RGB values and a float alpha value or a color name or a Color object")
        elif kwargs:
            if 'col_name' in kwargs:
                col_name = kwargs['col_name']
            if 'r' in kwargs:
                self.red = kwargs['r']
            if 'red' in kwargs:
                self.red = kwargs['red']
            if 'g' in kwargs:
                self.green = kwargs['g']
            if 'green' in kwargs:
                self.green = kwargs['green']
            if 'b' in kwargs:
                self.blue = kwargs['b']
            if 'blue' in kwargs:
                self.blue = kwargs['blue']
            if 'a' in kwargs:
                self.alpha = kwargs['a']
            if 'alpha' in kwargs:
                self.alpha = kwargs['alpha']

        if col_name is not None:
            try:
                web_color = webcolors.name_to_rgb(col_name)
                self.set_color(web_color.red, web_color.green, web_color.blue)
            except ValueError:
                raise ValueError(col_name + " is not a valid color name")

    def get_byte_representation(self) -> list:
        """
        Gets the RGBA values as list of ints from 0-255
        Returns:
            byte representation of color
        """
        r = self.red
        g = self.green
        b = self.blue
        a = round(255 * self.alpha)

        bytebuffer = list()
        bytebuffer.append(r)
        bytebuffer.append(g)
        bytebuffer.append(b)
        bytebuffer.append(a)

        return bytebuffer

    def __eq__(self, other):
        """deep equality check, by value of each RGBA value"""
        if not isinstance(other, self.__class__):
            return False

        return (self.red == other.red and
                self.green == other.green and
                self.blue == other.blue and
                self.alpha == other.alpha)



#!/usr/bin/env python
## package: bridges.base
from bridges.color import *

# from Element import Element

class LinkVisualizer():
    """
    @brief This class maintains the visual attributes of links that join bridges elements.
    
    Visual properties include color, thickness, and opacity.
    Objects of this class are stored as part of the Element class.
    Generally, a user will manipulate the LinkVisualizer returned from the
    Element's getLinkVisualizer(Element it) method (which it is the bridges element
    this element is linked to), and then set attributes using its methods. Links are
    utilized in all types of linked lists, tree and graph structures.
    
    Supported attribute values are as follows:<p>
    
    <b>Supported Colors by name:</b> CSS colors; see the Color class for full list, or  at
    https://drafts.csswg.org/css-color-3/#svg-color</b>: <p>
    
    <b> Color by RGBA Specification :</b>  Range: 0-255 for each component <p>
    
    <b> Thickness: </b> Range : 0.0-50.0
    
    <b> Opacity: </b> Range (0.0-1.0) </p>
    
    @author Mihai Mehedint, Kalpathi Subramanian, Matthew McQuaigue
    
    @date 2018, 6/24/19
    
    \sa Example Tutorial at <br>
    https://bridgesuncc.github.io/tutorials/SinglyLinkedList.html
    """

    def __init__(self) -> None:
        """
        Constructor for the link visualizer'
        Retruns:
            None
        """
        self._color = Color(70, 130, 180, 1.0)
        self._thickness = 1.0
        self._label = ""

    @property
    def thickness(self) -> float:
        """
        Getter for the thickness of links visualization
        Returns:
            float: the thickness
        """
        return self._thickness

    @thickness.setter
    def thickness(self, th):
        """
        Setter for the thickness of links in visualization
        Args:
            (float) th: the thickness to be applied to the link
        Returns:
            None
        """
        self._thickness = th

    @property
    def color(self) -> Color:
        """
        Getter for the color of the link in the visualization
        Returns:
            Color
        """
        return self._color

    @color.setter
    def color(self, *args, **kwargs):
        """
        Setter for the color of the element in the bridges visualization
        Args:
            (optional) list: requires either 3 ints 0-255 for RGB and an optional 
            float 0.0-1.0 for alpha EX: color = [0, 255, 0, 1.0].
            (optional) str: string representing the element color. Supported web colors: 
            https://developer.mozilla.org/en-US/docs/Web/CSS/color_value
        Returns:
            None
        Raises:
            ValueError: if the color name provided is not available
        """
        self._color = Color(*args, *kwargs)

    @property
    def opacity(self):
        """
        Getter for the element opacity
        Returns:
            opacity : element opacity 
        """
        return self.color.alpha

    @opacity.setter
    def opacity(self, opacity):
        """
        Setter for the elementopacity
        Args:
            opacity : opacity value (0-1.0) to set
        Returns:
            None
        """
        self.color.alpha = opacity

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, l):
        self._label = l

    def get_link_properties(self):
        """
        Getter for the link properties
        Returns:
            dict: link properties
        """
        link_props = {
            "color": [str(self.color.red), str(self.color.green), str(self.color.blue), str(self.color.alpha)],
            "thickness": str(self.thickness),
        }
        if self.label != "" or self.label is not None:
            link_props["label"] = self.label

        return link_props

    def get_label(self):
        """
        Getter for the link label
        Returns:
            string : link label
        """
        return self.label

    def set_label(self, label):
        """
        Setter for the element label
        Args:
            label : link label (string)
        Returns:
            None
        """
        self.label = label




import json
import os
import pickle

'''
@brief This class is used to cache very large datasets, such as the OpenStreet Map data.
This is for internal use only for efficiency.

@author Jay Strahler, Erik Saule

@date  2019, 2020

'''

class lru_cache():

    def __init__(self, max_cache_size: int = 0):
        self.max_cache_size = max_cache_size + 1
        self.lru = []

        if "HOME" in os.environ:
            self.cache_dir = os.environ['HOME'] + "/.cache/bridges_data/python/"

        elif os.name == "nt" and os.path.isdir(os.getenv('APPDATA') + "/local/bridges_data/python/"):
            self.cache_dir = os.getenv('APPDATA') + "/local/bridges_data/python/"

        else:
            self.cache_dir = os.getcwd() + "/.cache/bridges_data/python/"


        if not os.path.isdir(self.cache_dir):
            os.makedirs(self.cache_dir)


    def put(self, hash, content):
        try:
            with open(self.cache_dir + "/lru.txt", "rb") as fp:
                self.lru = pickle.load(fp)
        except:
            pass

        try: # Removes the location requested by the user from the LRU list
            self.lru.remove(hash)
        except:
            pass

        self.lru.insert(0, hash)

        # Removes least used map if there are more than 30 maps already saved
        if (len(self.lru) >= self.max_cache_size):
            re = self.lru[len(self.lru)-1]
            if (os.path.isfile(self.cache_dir + re)):
                os.remove(self.cache_dir + re)
            self.lru.remove(re)

        with open(self.cache_dir + "lru.txt", "wb") as fp:   #Pickling
            pickle.dump(self.lru, fp)

        with open(self.cache_dir + hash, "wb") as f:   # write to file in cache
            pickle.dump(content, f)
        return


    def get(self, hash):
        with open(self.cache_dir + hash, "rb") as j:
            try:
                data = pickle.load(j)
            except:
                print("Error: Issue reading locally cached file")
        return data



    def inCache(self, file_name):
        if (os.path.isfile(self.cache_dir + file_name)):
            return True
        return False



#!/usr/bin/env python
from bridges.tree_element import *


class BinTreeElement(TreeElement):
    """
    @brief This class is extended from the TreeElement class  and can be used to create binary tree element objects.
    
    The BinTree element class is the building block for creating binary tree structures.
    It contains two children (viz., left, right).
    
    BinTreeElement contains a visualizer (ElementVisualizer) object for setting visual
    attributes (color, shape, opacity, size), necessary for displaying them in a
    web browser.
    
    Elements also have a LinkVisualizer object, that is used when they are linked to
    another element, appropriate for setting link attributes, for instance, between
    the current element and its left or  right child
    
    @author Kalpathi Subramanian, Mihai Mehedint, Matthew McQuaigue
    
    @date   2018, 7/23/19, 1/6/21
    
    \sa Binary tree tutorial, https://bridgesuncc.github.io/tutorials/BinTree.html
    """

    def __init__(self, **kwargs) -> None:
        """
        Constructor for an empty Binary Tree Element
        kwargs:
            label: The label for the tree element that is displayed in the visualization
            e: the generic object that the binary tree element will hold
            left: the binary tree element assigned to child 0
            right: the binary tree element assigned to child 1
        Returns:
            None
        """
        if 'e' in kwargs:
            if 'e' in kwargs and 'label' in kwargs:
                super(BinTreeElement, self).__init__(label=kwargs['label'], e=kwargs['e'])
            else:
                super(BinTreeElement, self).__init__(e=kwargs['e'])
        else:
            super(BinTreeElement, self).__init__()
        if 'left' in kwargs and 'right' in kwargs:
            super(BinTreeElement, self).add_child(kwargs['left'])
            super(BinTreeElement, self).add_child(kwargs['right'])
        else:
            super(BinTreeElement, self).add_child(None)
            super(BinTreeElement, self).add_child(None)


    def get_data_structure_type(self) -> str:
        """
        Get the data structure type
        Returns:
            str: representing the data structure type
        """
        return "BinaryTree"

    @property
    def left(self):
        """
        Getter for the left element for the binary tree
        Returns:
            TreeElement: left child of this element
        """
        return self.get_child(0)

    @left.setter
    def left(self, l) -> None:
        """
        Setter for the left element of a binary tree
        Args:
            l: the left element to set
        Returns:
            None
        """
        self.set_child(0, l)

    @property
    def right(self):
        """
        Getter for the right element for the binary tree
        Returns:
            TreeElement: the right child of this element
        """
        return self.get_child(1)

    @right.setter
    def right(self, r) -> None:
        """
        Setter for the right element of a binary tree
        Args:
            r: the right element to set
        Returns:
            None
        """
        self.set_child(1, r)



#!/usr/bin/env python
from bridges.element import *
import json

class TreeElement(Element):
    """ 
    @brief This class extends Element to represent general trees with
    arbitrary number of children.
    
    TreeElement nodes can have an arbitrary number of child nodes(held in
    in a vector in the order in which they were added). The
    visualization of trees assumes that the children are drawn in order
    from left to right.
    
    Tree Elements have labels (string) that are displayed on the visualization.
    Elements take an generic object E as a user defined parameter, which can be
    any native type or object.
    
    Elements contain a visualizer (ElementVisualizer) object for setting visual
    attributes (color, shape, opacity, size), necessary for displaying them in a web
    browser.
    
    Elements also have a LinkVisualizer object that is used when they are
    linked to another element, appropriate for setting link attributes, between parent
    and child nodes.
    
    @author Matthew McQuaigue
    
    @date   2018, 7/23/19
    \sa Kd tree tutorial https://bridgesuncc.github.io/tutorials/KdTree.html
    """

    QUOTE = "\""
    COMMA = ","
    COLON = ":"
    OPEN_CURLY = "{"
    CLOSE_CURLY = "}"
    OPEN_PAREN = "("
    CLOSE_PAREN = ")"
    OPEN_BOX = "["
    CLOSE_BOX = "]"

    def __init__(self, **kwargs) -> None:
        """
        Constructor for Tree Element
        Kwargs:
            (0): The label for the tree element that shows in visualization
            (1): the generic object 'e' that the tree element will hold
            (2): the tree element assigned to child 0
            (3): the tree element assigned to child 1
        Returns:
            None
        """
        self.children = []
        if 'label' in kwargs and 'e' in kwargs:
            super(TreeElement, self).__init__(label = kwargs['label'], val = kwargs['e'])
        elif 'e' in kwargs:
            super(TreeElement, self).__init__(val = kwargs['e'])
        else:
            super(TreeElement, self).__init__()
        if 'left' in kwargs:
            self.children.append(kwargs['left'])
        if 'right' in kwargs:
            self.children.append(kwargs['right'])


    def get_data_structure_type(self) -> str:
        """
        Gets the data structure type
        Returns:
            str: representing the data structure type
        """
        return "Tree"

    def add_child(self, child) -> None:
        """
        Adds a child to this parent node
        Args:
            child: the child node to add
        Returns:
            None
        """
        self.children.append(child)

    def get_number_of_children(self) -> int:
        """
        Get the number of children at this node
        Returns:
            int: representing the number of children
        """
        return len(self.children)

    def set_child(self, index: int, child) -> None:
        """
        Adds a child to the node that will be added at the next open position
        Args:
            index: index to add child
            child: child to add to tree
        Returns:
            None
        Raises:
            ValueError: If the index is higher than the number of children
        """
        if index < len(self.children):
            self.children[index] = child
        else:
            raise ValueError("Index is higher than number of children")

    def get_child(self, index: int):
        """
        Gets a child at particular index
        Args:
            index: index to get child
        Returns:
            the child element at this index
        Raises:
            ValueError: if the index is higher than the number of children
        """
        if index < len(self.children):
            return self.children[index]
        else:
            raise ValueError("Index is higher than number of children")

    def get_data_structure_representation(self) -> dict:
        """
        Get the hierarchical JSON of the tree representation (internal use only)
        Returns:
            dict: representing the data structures json 
        """
        # json_dict = {
        #     "nodes": self._pre_order(self)
        #
        json_dict = ""
        json_dict += "{" + self.QUOTE + "nodes" + self.QUOTE + ":" + "{" + str(self._pre_order(self)) + "}" + "}"
        return json.loads(json_dict)

    def _pre_order(self, root) -> dict:
        """
        Use a preoreder traversal to directly extract a hierarchical
        JSON representation of the tree
        Args:
            root: the root of the tree structure
        Returns:
            dict: representing the json to be returned
        """
        json_str = ""
        children = ""
        link_props = ""
        elem_rep = ""
        t_str = str()
        num = root.get_number_of_children()
        if root is not None:
            elem_rep = root.get_element_representation()
            elem_rep = json.dumps(elem_rep)
            t_str = elem_rep[1:-1]
            json_str = t_str
            if root.get_number_of_children() > 0:
                json_str += "," + self.QUOTE + "children" + self.QUOTE + ":" + "["
            for k in range(0, root.get_number_of_children()):
                if root.get_child(k) is None:
                    json_str += "{" + self.QUOTE + "name" + self.QUOTE + ":" + \
                        self.QUOTE + "NULL" + self.QUOTE + "}" + ","
                else:
                    lv = root.get_link_visualizer(root.get_child(k))
                    json_str += "{"
                    if lv is not None:
                        json_str += self.QUOTE + "linkProperties" + self.QUOTE + ":" + "{" + \
                            self.QUOTE + "color" + self.QUOTE + ":" + \
                            "[" + \
                            str(lv.color.red) + "," + \
                            str(lv.color.green) + "," + \
                            str(lv.color.blue) + "," + \
                            str(lv.color.alpha) + "]" + "," + \
                            self.QUOTE + "thickness" + self.QUOTE + ":" + \
                            str(lv.thickness) + "}" + ","
                    else:
                        json_str += "linkProperties" + ":" + "{}" + ","

                    json_str += str(self._pre_order(root.get_child(k)))
                    json_str += "}" + ","
            if len(json_str) > 0:
                json_str = json_str[0:-1]
            json_str += "]"
        return json_str

        # k = 0
        # json_str = dict()
        # if root is not None:
        #     #  first get the node representationS
        #     elem_rep = root.get_element_representation()
        #     if root.get_number_of_children() > 0:
        #         elem_rep["children"] = []
        #     for i in range(0, root.get_number_of_children()):
        #         if root.get_child(i) is None:
        #             temp = {
        #                 "name": "NULL"
        #             }
        #             elem_rep["children"].append(temp)
        #         else:
        #             lv = root.get_link_visualizer(root.get_child(i))
        #             if lv is not None:
        #                 temp = {
        #                     "linkProperties": {
        #                         "color": [lv.color.red, lv.color.green, lv.color.blue, lv.color.alpha],
        #                         "thickness": lv.thickness
        #                     }
        #                 }
        #                 elem_rep["children"].append(temp)
        #             else:
        #                 temp = {
        #                     "linkProperties": {}
        #                 }
        #                 elem_rep["children"].append(temp)
        #             elem_rep['children'].append(self._pre_order(root.get_child(i)))
        #     json_str = elem_rep
        #
        # return json_str



from bridges.array import *

class Array2D(Array):
    """
    brief This class can be used to create 2D arrays of type Element<E>.
    
    @author 	Kalpathi Subramanian, Matthew McQuaigue
    
    @date  	10816, 51717, 53018
    
    This class can be used to create arrays of type Element<E>  where E
    is a generic object representing application specific data.
    
    Arrays are internally represented as 1D arrays; currently 1D, 2D  and
    3D arrays are supported.
    
    Example Tutorial at: https//:bridgesuncc.github.io/tutorials/Array.html 
    (1D, 2D, and 3D Array)<br>
    """

    def __init__(self, rows=None, cols=None):
		##
		# @brief Create an array object with the specified dimensions
		# @param rows  number of rows in array (int)
        # @param cols  number of cols in array (int)
        super(Array2D, self).__init__()
        if rows is not None and cols is not None:
            dim = [cols, rows, 1]
            self.num_rows = rows
            self.num_cols = cols
            self.set_size(2, dim)
        else:
            self.size = 0
            self.num_rows = 0
            self.num_cols = 0



from bridges.link_visualizer import *

class Edge():
    """
    @brief This class is used to represent the edges in a graph and will appear as links in the BRIDGES graph visualization.

    This object is used in graphs and graph algorithms such as DFS, BFS and shortest
    path algorithms that need to visit graph edges. The adjacency list
    representation uses them as the generic paramter, as SLelement<Edge>
    bridges represents Edges as links between pairs of elements

    @author Matthew McQuaigue, Kalpathi Subramanian

    @date 2019, 2020, 2021
    """


    def __init__(self, v1, v2, data=None, label: str = None, color: Color = None, thickness: float = None) -> None:
        """
        @brief Constructor for a edge
        Args:
            v1: first vertex of the edge
            v2: second vertex of the edge
            data: the data the edge will hold
            label: edge label
            color: edge color
            thickness: edge thickness
        """
        self._from_vertex = v1
        self._to_vertex = v2
        self._edge_data = data
        self._lvis = LinkVisualizer()

        if label:
            self.label = label

        if color:
            self.color = color

        if thickness:
            self.thickness = thickness

    @property
    def tov(self):
        """
        Getter for the to vertex
        Returns:
            vertex: terminating vertex of edge 
        """
        return self._to_vertex

    @property
    def fromv(self):
        """
        Getter for the from vertex
        Returns:
            vertex : source vertex  of edge
        """
        return self._from_vertex

    @property
    def destination(self):
        """
        Getter for the to vertex
        Returns:
            vertex: terminating vertex of edge
        """
        return self._to_vertex

    @property
    def source(self):
        """
        Getter for the from vertex
        Returns:
            vertex : source vertex  of edge
        """
        return self._from_vertex

    @property
    def thickness(self) -> float:
        """
        Getter for the link thickness
        Returns:
             float : link thickness (1.0-10.0 range)
        """
        return self._lvis.thickness

    @thickness.setter
    def thickness(self, th: float) -> None:
        """
        Setter for the thickness of edge
        Args:
            th: thickness to be applied (1.0-10.0 range)
        Returns:
            None
        """
        self._lvis.thickness = th

    @property
    def edge_data(self):
        """
        Getter for edge data
        Returns:
             str : data associated with this edge (generic object)
        """
        return self._edge_data

    @edge_data.setter
    def edge_data(self, data) -> None:
        """
        Setter for edge data
        Args:
            data: data for the edge to hold (generic object)
        Returns:
            None
        """
        self._edge_data = data

    @property
    def color(self):
        """
        Getter for edge color
        Returns:
                color of edge (see link visualizer class for setting options
        """
        return self._lvis.color

    @color.setter
    def color(self, color):
        """
        Setter for edge color
        Args:
            color : color to be set (see link visualizer class for setting options)
        Returns:
            None
        """
        self._lvis.color = color

    @property
    def opacity(self):
        """
        Getter for the edge opacity
        Returns:
            opacity : edge opacity
        """
        return self.color.alpha

    @opacity.setter
    def opacity(self, opacity):
        """
        Setter for the edge opacity
        Args:
            opacity : opacity value (0-1.0) to set
        Returns:
            None
        """
        self.color.alpha = opacity

    @property
    def label(self):
        """
        Getter for the edge label
        Returns:
            (string) : edge label
        """
        return self._lvis.label

    @label.setter
    def label(self, l):
        """
        Setter for the edge label
        Args:
            l : lavel value (string) to set
        Returns:
            None
        """
        self._lvis.label = l

    def get_edge(self):
        """
        Get this edge object
        Returns:
            self: this edge object
        """
        return self



from bridges.graph_benchmark import *
from bridges.line_chart import *
from datetime import datetime
import time as time_


class BFSBenchamrk(GraphBenchmark):
    """
    @brief Benchmarks Breadth First Search algorithms.

	Benchmarks BFS algorithms and add time series to a LineChart.

	One can also set a maximum time spent on a particular run
	using setTimeCap().

	The BFS algorithms must have for prototype:

	and can be passed to the run function for being
	benchmarked. A typical use would look something like

	code
	lc LineChart();
	sb BFSBenchmark(lc);
	sb.run("mybfsalgorithm", bfsalgo);
	endcode

	@author Erik Saule
	@date 07212019
    """
    def __init__(self, p):
        super(BFSBenchamrk, self).__init__()
        self.__plot = p
        self.__plot.x_label = "Number of Edges"
        self.__plot.y_label = "Runtime (in s)"

    def run(self, algo_name, bfsalgo):
        """
        benchmark one implementation
        Args:
            algo_name: screen name of the algorithm to be used in the visualization
            bfsalgo: pointer to the sorting function to benchmark
        """
        time = []
        vtx_count = []
        edge_cnt = []

        years = 0
        while years < 120:
            year = 2019 - years
            graph = GraphAdjList()
            vertex_count, edge_count = self._genertate_wiki_data_movie_actor(int(year), 2019, graph)
            root = self._highest_degree_vertex(graph)
            level = dict()
            parent = dict()
            start = float((time_.time() * 1000))
            bfsalgo(graph, root, level, parent)
            end = float((time_.time() * 1000))
            elapsed_time = end - start
            time.append(elapsed_time)
            vtx_count.append(vertex_count)
            edge_cnt.append(edge_count)

            if elapsed_time > self.time_cap:
                break
            years = 1.2 * years + 1

        self.__plot.set_x_data(algo_name, edge_cnt)
        self.__plot.set_y_data(algo_name, time)






import random
import time as time_


class SortingBenchmark:
    """
    @brief Benchmarks sorting algorithm

	Benchmarks sorting algorithms and add time series to a LineChart.

	The benchmark goes from an initial size controlled by
	setBaseSize() to a largest size controlled by setMaxSize(). One
	can also set a maximum time spent on a particular run using
	setTimeCap().

	The benchmark goes from a array size of n to the next one of
	geoBase  n + increment, where the base is controlled by
	geometric and increment is controlled by
	increment. For simpler use one can set a purley linear
	sampling with linear_range() or a purely geometric one with
	geometric_range().

	sorting algorithms can be passed to the run function for being benchmarked. A typical use would look something like

	\code{.py}
	def mysort(int array, int arraysize)
	lc = LineChart();
	sb SortingBenchmark(lc);
	sb.linear_range (100, 1000, 5);
	sb.run("mysortingalgorithm", mysort);
	\endcode

	@author Erik Saule
	@date 07/20/2019
    """
    def __init__(self, p):
        p.x_label = "Size of Array"
        p.y_label = "Runtime (in us)"
        self._plot = p
        self.r = random
        self._max_size = 1
        self._base_size = 1
        self._increment = 1
        self._geo_base = 1
        self._time_cap_ms = float('inf')
        self.debug = False

    @property
    def plot(self):
        return self._plot

    @plot.setter
    def plot(self, p):
        self._plot = p

    @property
    def max_size(self):
        """
        Getter for the max_size of array
        Returns:
            the max size
        """
        return self._max_size

    @max_size.setter
    def max_size(self, size):
        """
        Puts a cap on the largest array to be used
        Args:
            size: Maximum size considered
        """
        self._max_size = size

    @property
    def base_size(self):
        return self._base_size

    @base_size.setter
    def base_size(self, size):
        """
        Smallest array to be used
        Args:
            size: size of the smallest array to use
        """
        self._base_size = size

    @property
    def increment(self):
        return self._increment

    @increment.setter
    def increment(self, inc):
        """
        Sets the increment for the benchmark size
        Args:
            inc: new value of the increment
        """
        self._increment = inc

    @property
    def geometric(self):
        return self._geo_base

    @geometric.setter
    def geometric(self, base):
        """
        Sets a geometric progression for the benchmark size
        Args:
            base: new base of the geometric progression
        """
        self._geo_base = base

    @property
    def time_cap(self):
        return self._time_cap_ms

    @time_cap.setter
    def time_cap(self, cap_in_ms):
        """
        Sets an upper bound to the time of a run.
        The benchmark will end after a run if it takes more than the
        given amount of time. So it is possible a particular run takes
        more than the alloted time, but that will be the last run.
        Args:
            cap_in_ms: time limit in seconds
        """
        self._time_cap_ms = cap_in_ms

    def linear_range(self, base_sz, max_sz, nb_point):
        """
        brief The benchmark will sample a range with a fixed number of
		points.
		The benchmark will sample about nbPoint equally distributed in
		the range [base_size, max_size]
        Args:
            base_sz: lower bound of the range sampled
            max_sz: upper bound of the range sampled
            nb_points: number of sample
        """
        self.base_size = base_sz
        self.max_size = max_sz
        self.increment = ((max_sz - base_sz) / nb_point)
        self.geometric = 1.0

    def geometric_range(self, base_sz, max_sz, base):
        """
        The benchmark will sample a range using in geometrically
		increasing sequence.
        The benchmark will sample the range [baseSize; maxSize] using a
		geometric distribution in base base. That is to say, it will
		sample base_ize, base*base_ize, base*base*base_ize, ...
        Args:
            base_sz: lower bound of the range sampled
            max_sz: upper bound of the range sampled
            base: base of the geometric increase
        """
        self.base_size = base_sz
        self.max_size = max_sz
        self.increment = 0
        self.geometric = base
        if base <= 1.0:
            print("base should be > 1.0")

    def __generate(self, arr, n):
        for i in range(0, n):
            arr.append(self.r.randint(0, 2*n))

    def __check(self, arr, n):
        ok = True
        for i in range(1, n):
            if arr[i] < arr[i -1]:
                ok = False
                break
        return ok

    def run(self, algo_name, runnable):
        """
        benchmark one implementation
        Args:
            algo_name: screen name of the algorithm to be used in the visualization
            runnable: pointer to the sorting function to benchmark
        :return:
        """
        time = []
        x_data = []

        if self.debug:
            print(self.geometric)
            print(self.increment)

        n = self.base_size
        while n < self.max_size:
            arr = []
            self.__generate(arr, int(n))

            start = int(round(time_.time() * 1000))
            runnable(arr)
            end = int(round(time_.time() * 1000))
            runtime = end - start

            if self.__check(arr, int(n)) == False:
                print("Sorting algorithm " + algo_name + " is incorrect")

            time.append(float(runtime))
            x_data.append(float(n))

            if runtime > self.time_cap:
                break

            n = max(self.geometric * n + self.increment, n + 1)
        self.plot.set_x_data(algo_name, x_data)
        self.plot.set_y_data(algo_name, time)



from bridges.bridges import *
from bridges.game_grid import *
from bridges.socket_connection import *
from abc import ABC, abstractmethod


class GameBase(ABC):
    """
    @brief This is the base class for all games in BRIDGES

    @author David Burlinson, Erik Saule
    @date 2020, 2021

    """
    debug = True

    def __init__(self, assid, login, apikey, rows, cols):
        """
        PROTECTED constructor prevents the object from being directly created. Since GameBase is meant to be a purely internal class, that seems appropriate.

        Args:
            assid: assignment number
            login: user name
            apikey: authentication id
            rows: height of game grid
            cols: width of game grid
		Returns:
			None
        """
        self.game_base_init(assid, login, apikey, rows, cols)
        self.grid_state = dict

    def game_base_init(self, id, log, key, rows, cols):
        """
        Initializes the gamebase object
        Args:
            id: assignment number
            log: user name
            key: authentication id
            rows: height of game grid
            cols: width of game grid
        """
        self.firsttime = True

        self.bridges = Bridges(id, log, key)

        #self.bridges.set_visualize_JSON(True)

        self.bridges.connector.set_server("games")

        self.grid = GameGrid(rows, cols)

        self.grid.set_encoding("rle")

        self.sock = SocketConnection(self.bridges)
        self.sock.setup_connection(log, id)

    def close(self):
        """
        close the socket connection (and game)
        """
        self.sock.close()
        
    def register_keypress(self, kl):
        self.sock.add_listener(kl)

    def start(self):
        """
        "starts the game. Called once at the beginning"
        """
        pass

    def initialize(self):
        """
        "Game initialization. . Called once at the beginning"
        """
        pass

    def game_loop(self):
        """
        "Method contains all of the game logica and can call other user defined methods"
        """
        pass

    def quit(self):
        """
        calling this function causes the game to end.

        Returns:
            None
        """
        self.game_started = False

    def set_title(self, title):
        """
        sets title of game

        Returns:
            None
        """
        self.bridges.set_title(title)

    def set_description(self, desc):
        """
        sets description of the game

        Returns:
            None
        """
        self.bridges.set_description(desc)

    def get_bg_color(self, row, col):
        """
        gets background color of a cell

        Args:
            row: the row of the cell
            col: the column of the cell

        Returns:
            returns a NamedColor
        """
        return self.grid.get_bg_color(row, col)

    def set_bg_color(self, row, col, color):
        """
        sets background color of a cell

        Args:
            row: the row of the cell
            col: the column of the cell
            color: color to be set

        Returns:
            None
        """
        self.grid.set_bg_color(row, col, color)

    def get_symbol(self, row, col):
        """
        gets symbol of the  cell at row, col

        Args:
            row: the row of the cell
            col: the column of the cell

        Returns:
            Symbol of type NamedSymbol
        """

        return self.grid.get_symbol(row, col)

    def get_symbol_color(self, row, col):
        """
        gets symbol color of the  cell at row, col

        Args:
            row: the row of the cell
            col: the column of the cell

        Returns:
            color of type NamedColor
        """

        return self.grid.get_symbol_color(row, col)

    def draw_symbol(self, row, col, s, c):
        """
        draw symbol s with color col at  cell (row, col)

        Args:
            row: the row of the cell
            col: the column of the cell
            s: symbol
            c: color

        Returns:
            color of type NamedColor
        """
        self.grid.draw_symbol(row, col, s, c)

    def render(self):
        """
        renders the board
        """

        if self.firsttime:
            self.firsttime = False

            self.bridges.set_data_structure(self.grid)
            try:
                self.bridges.visualize()
            except RuntimeError as e:
                print(e)

        self.grid_state = self.grid.get_data_structure_representation()
        self.sock.send_data(self.grid_state)

    @property
    def board_width(self):
        """
        setter/getter property for board width
        """
        dims = self.grid.dimensions
        return dims[1]

    @property
    def board_height(self):
        """
        setter/getter property for board height
        """
        dims = self.grid.dimensions
        return dims[0]




class reddit:

    def __init__(self, id = 0, title = "", author= "", score = 0, vote_ratio = 0, comment_count = 0, subreddit = "", post_time = 0, url = "", text = ""):
        self.id = id
        self.title = title
        self.author = author
        self.score = score
        self.vote_ratio = vote_ratio
        self.comment_count = comment_count
        self.subreddit = subreddit
        self.post_time = post_time
        self.url = url
        self.text = text

    @property
    def id(self):
        return self.__id

    @id.setter
    def id(self, i):
        self.__id = i

    @property
    def title(self):
        return self.__title

    @title.setter
    def title(self, i):
        self.__title = i

    @property
    def author(self):
        return self.__author

    @author.setter
    def author(self, i):
        self.__author = i

    @property
    def score(self):
        return self.__score

    @score.setter
    def score(self, i):
        self.__score = i

    @property
    def vote_ratio(self):
        return self.__vote_ratio

    @vote_ratio.setter
    def vote_ratio(self, i):
        self.__vote_ratio = i

    @property
    def comment_count(self):
        return self.__comment_count

    @comment_count.setter
    def comment_count(self, i):
        self.__comment_count = i

    @property
    def subreddit(self):
        return self.__subreddit

    @subreddit.setter
    def subreddit(self, i):
        self.__subreddit = i

    @property
    def post_time(self):
        return self.__post_time

    @post_time.setter
    def post_time(self, i):
        self.__post_time = i

    @property
    def url(self):
        return self.__url

    @url.setter
    def url(self, i):
        self.__url = i

    @property
    def text(self):
        return self.__text

    @text.setter
    def text(self, i):
        self.__text = i


from bridges.connector import *
from bridges import ColorGrid
import os

import json

class Bridges:
    """
    @brief The bridges class is the main class that provides interfaces to datasets, maintains user and assignment information, and connects to the bridges server.
    
    The bridges class is responsible  for initializing the bridges system, specifying
    parameters (user id, assignment id, title, description, data structure
    type, etc) for the student assignment, generating the data structure representation
    and transmission to the bridges server. In addition, it provides interfaces to
    a number of real-world datasets, that makes it easy to access the data for use
    algorithms/data structure assignments. <br>
    
    Datasets. The datasets that are currently supported through the BRIDGES API
    include USGS Earthquake Data, IMDB Actor/Movie Data (2 versions), Gutenberg Book
    Collection Meta Data, a Video Game Dataset, Shakespeare Dataset, OpenStreetMap and Elevation map data. More information
    is found in the respective methods (below) and at 
    https://bridgesuncc.github.io/datasets.html <p>
    
    A typical bridges program includes creating the bridges object, followed by creation
    of the data structure by the user, assigning visual attributes to elements of the
    data structure, followed by specification of teh data structure type  and the
    call to visualize the data structure (bridges::setDataStructure() and visualize()
    methods).
    
    @author Sean Gallagher, Kalpathi Subramanaian, Mihai Mehedint, David Burlinson, Matthew Mcquaigue, Erik Saule, Jamie Payton
    
    @date 2015, 2016, 2017, 2018, 2019, 2020, 2021
    """

    _MaxTitleSize = 50
    _MaxDescSize = 250
    _projection_options = {"cartesian", "albersusa", "equirectangular", "window"}

    @property
    def window(self) -> [float]:
        """
        This function returns the current window size.  This only works for graph data types. And the coordinate system need to be set to "window" using set_coord_system_type(), setting this value will set "window" for you.
        Returns:
           return a list of 4 floats [x1, x2, y1, y2]
        """
        return self._window

    @window.setter
    def window(self, value: [float]) -> None:
        """
        This function sets the current window size that will rendered by default in the view.  This only works for graph data types. And the coordinate system need to be set to "window" using set_coord_system_type(), setting this value will set "window" for you.
        """
        try:
            new_window = [float(x) for x in value]
        except ValueError:
            raise ValueError("Value for window should be a list of 4 numbers")
        except TypeError:
            raise TypeError("Value for window should be a list of 4 numbers")

        self.set_coord_system_type("window")
        self._window = new_window

    def __init__(self, assignment, username=None, appl_id=None):
        """
        Bridges constructor
        Args:
            (int) assignment: the number your bridges assignment will have
            (str) username: your bridges username, optional if BRIDGES_USER_NAME in env
            (str) appl_id: your appl authentication key from bridges acc, optional if BRIDGES_API_KEY in env
        Returns:
            None
        """
        username = username if username is not None else os.getenv("BRIDGES_USER_NAME")
        appl_id = appl_id if appl_id is not None else os.getenv("BRIDGES_API_KEY")

        self._assignment_part = 0
        self._assignment = 0
        self._username = str()
        self._key = str()
        self._title = str()
        self._description = str()
        self.set_assignment(assignment)
        self.set_username(username)
        self.set_key(appl_id)

        self.connector = Connector(self.get_key(), self.get_username(), self.get_assignment())
        self._coord_system_type = "cartesian"
        self._json_flag = False
        self._post_url_flag = True
        self._map_overlay = False
        self._map = ["us", "all"]
        self._window = [0.0, 0.0, 0.0, 0.0]
        self.ds_handle = None
        self.vis_type = ""

    def set_data_structure(self, ds):
        """
        Set the data structure type.  This method sets the handle to the current data structure; this can be an array, the head of a linked list, root of a tree structure, a graph Arrays of upto 3 dimensions are suppported. It can be any of the data structures supported by BRIDGES. Polymorphism and type casting is used to determine the actual data structure and extract its representtion.
        Args:
            ds: the data structure to visualize
        Returns:
            None
        Raises:
            ValueError: if it is not a BRIDGES data structure
        """
        try:
            self.ds_handle = ds
            self.vis_type = ds.get_data_structure_type()
        except ValueError:
            print("Exception Thrown: Data structure passed to BRIDGES is null!\n")


    def set_visualize_JSON(self, flag):
        """
        This method controls if the data structure's JSON representation is printed to the console or not
        Args:
            flag : flag that controls if the JSON of the data structure representation is output to console
        Returns:
            None
        """
        self._json_flag = flag

    def post_visualization_link(self, flag):
        """
        This method controls (with a flag) if the visualization url is printed to the console or not.
        Args:
            flag : flag that controls if the url to the visualization is output to console
        """
        self._post_url_flag = flag

    def visualize(self) -> None:
        """
        Method for generating the representation of the data structure in the form of JSON and sends the information to the bridges server for generating the visualization
        Returns:
            None
        """
        nodes_links_str = ""

        if self.vis_type == "Tree" or self.vis_type == "BinaryTree" or self.vis_type == "AVLTree" or\
                self.vis_type == "SinglyLinkedList" or self.vis_type == "DoublyLinkedList" or \
                self.vis_type == "MultiList" or self.vis_type == "CircularSinglyLinkedList" or \
                self.vis_type == "CircularDoublyLinkedList" or self.vis_type == "Array" or \
                self.vis_type == "GraphAdjacencyList" or self.vis_type == "ColorGrid" or self.vis_type == "GraphAdjacencyMatrix" or \
                self.vis_type == "largegraph" or self.vis_type == "KdTree" or self.vis_type == "SymbolCollection" or \
                self.vis_type == "GameGrid" or self.vis_type == "BinarySearchTree" or self.vis_type == "LineChart" or \
                self.vis_type == "Audio" or self.vis_type == "SymbolCollectionV2" or self.vis_type == "Scene":
            nodes_links_str = self.ds_handle.get_data_structure_representation()

        ds = {
            "visual": self.vis_type,
            "title": self._title,
            "description": self._description,
            "coord_system_type": self._coord_system_type,
            "map_overlay": self._map_overlay,
            "map": self._map,
        }
        if self.window is not None and len(self.window) == 4:
            ds['window'] = self.window

        ds.update(nodes_links_str)

        ds_json = json.dumps(ds)
        if self._json_flag:
            print(ds_json)

        response = self.connector.post("/assignments/" + self.get_assignment(), ds_json)

        if response == 200 and self._post_url_flag:
            print("\nCheck Your Visualization at the following link:\n\n" +
                  self.connector.get_server_url() + "/assignments/" + str(self._assignment) +
                  "/" + self._username + "\n\n")

            self._assignment_part = self._assignment_part + 1


    def set_assignment(self, assignment):
        """
        Setter for assignment id (must be positive)
        Args: 
           assignment: assignment number to be set
        Returns:
           None
        """
        force = os.getenv("FORCE_BRIDGES_ASSIGNMENT", "")
        if (force != ""):
            assignment = int(force)
        if assignment < 0:
            ValueError("Assignment value must be >= 0")
        elif self._assignment >= 0:
            self._assignment_part = 0
        self._assignment = assignment

    def get_assignment(self) -> str:
        """
        Getter for the assignment id
        Returns:
            str: representing the full assignment id including subassignment aspect
        """
        if self._assignment_part < 10:
            return str(self._assignment) + ".0" + str(self._assignment_part)
        else:
            return str(self._assignment) + "." + str(self._assignment_part)

    def set_title(self, title) -> None:
        """
        Setter for the title of the bridges visualization
        Args:
            (str) title: representing the title
        Returns:
            None
        """
        if len(title) > self._MaxTitleSize:
            print("Visualization Title restricted to" + str(self._MaxTitleSize) + " characters." + " truncated title...")
            self._title = title[:self._MaxTitleSize]
        else:
            self._title = title

    def set_description(self, description) -> None:
        """
        Setter for the description of the bridges visualization
        Args:
            (str) description: representing the assignment description
        Returns:
            None
        """
        if len(description) > self._MaxDescSize:
            print("Visualization Description restricted to " + str(self._MaxDescSize) + " Truncating description..")
            self._description = description[0:self._MaxDescSize]
        else:
            self._description = description

    def set_map_overlay(self, flag):
        """
        Setter for if the visualization will have a map overlay
        Args:
            (bool) flag: boolean for if map overlay
        Returns:
            None
        """
        self._map_overlay = flag

    @property
    def map(self):
        return self._map

    @map.setter
    def map(self, new_map):
        self._map = new_map

    def set_coord_system_type(self, coord):
        """
        Setter for the coordinate system type to use in the visualization
        Args:
           coord: coordinate system type (used in map overlays (can be 
           "cartesian", "albersusa", "equirectangular")
        """
        if coord in self._projection_options:
            self._coord_system_type = coord
        else:
            print("Unrecognized coordinate system \'" + coord + "\', defaulting to cartesian. Options:")
            self._coord_system_type = "cartesian"

    def get_color_grid_from_assignment(self, user: str, assignment: int, subassignment: int = 0) -> ColorGrid:
        """
        Reconstruct a ColorGrid from an existing ColorGrid on the bridges server
        Args:
            user(str): the name of the user who uploaded the assignment
            assignment(int): the ID of the assignment to get
            subassignment(int): the ID of the subassignment to get (default 0)
        Returns:
            ColorGrid: the ColorGrid stored in the bridges server
        """
        from bridges.data_src_dependent.data_source import get_color_grid_from_assignment
        return get_color_grid_from_assignment(self.connector.server_url, user, assignment, subassignment)

    def set_username(self, username):
        """
        Setter for username (must be a string)
        Args: 
           username: username to be set
        Returns:
           None
        """
        force = os.getenv("FORCE_BRIDGES_USERNAME", "")
        if (force != ""):
            username = force
        self._username = username.replace(" ", "+")
    
    def get_username(self):
        """
        Getter for the assignment user name (BRIDGES credentials)
        Returns:
            str: user name
        """
        return self._username

    def get_assignment_id(self):
        """
        Getter for the assignment number
        Returns:
            int: assignment number
        """
        return self._assignment


    def set_key(self, apikey):
        """
        Setter for API Key (BRIDGES Credentials)
        Args: 
           apikey: api key to be set
        Returns:
           None
        """
        force = os.getenv("FORCE_BRIDGES_APIKEY", "")
        if (force != ""):
            apikey = force
        self._key = apikey.replace(" ", "+")

    def get_key(self):
        """
        Getter for the API key (BRIDGES credentials)
        Returns:
            str: user's API key
        """
        return self._key

    def set_server_url(self, server_url: str) -> None:
        """
        Sets url for the output of the visualize method
        Args:
            server_url: string, must be one of {live, clone, local, games}
		Returns:
            None
        """
        self.connector.set_server_url(server_url)





import json

class SymbolCollection:
    """
    @brief the SymbolCollection object is a container object to hold a collection of symbols
    (rectangles, polygons, circles, labels).
    
    @author Matthew Mcquaigue
    
    @date 2018, 7/23/19
    
    \sa Symbol Collection tutorial, 
    https://bridgesuncc.github.io/tutorials/Symbol_Collection.html
    """
    def __init__(self):
        """
        Constructor for collection of symbols
        """
        self._symbols = []
        self._domainxmin = -100
        self._domainxmax = 100
        self._domainymin = -100
        self._domainymax = 100
        self._autoupdateviewport = True

    def setviewport(self, xmin, xmax, ymin, ymax):
        """
        Set the display size.
        Args:
            xmin: min x coordinate
            xmax: max x coordinate
            ymin: min y coordinate
            ymax: max y coordinate
        """
        self._autoupdateviewport = False
        self._domainxmin = xmin
        self._domainxmax = xmax
        self._domainymin = ymin
        self._domainymax = ymax    
        
    def get_data_structure_type(self):
        """
        Getter for the data structure type as a JSON string
        Returns
            str: the data structure type
        """
        return "SymbolCollectionV2"

    def add_symbol(self, s):
        """
        Add a symbol to the collection
        Args:
             s : symbol to be added
        Returns:
             none 
        """
        self._symbols.append(s)

    def update_axis_domains(self, s):
        """
        Update the bounding box of the symbol collection
        Args:
             s: input symbol to update dimensions
        Returns:
             none
        """
        # dims = s.get_dimensions()
        #
        # if dims[0] < self._domainxmin:
        #     self._domainxmin = dims[0]
        # if dims[1] > self._domainxmax:
        #     self._domainxmax = dims[1]
        #
        # if dims[2] < self._domainymin:
        #     self._domainymin = dims[2]
        # if dims[3] > self._domainymax:
        #     self._domainymax = dims[3]
        return

    def get_data_structure_representation(self):
        """
        Getter for the data structure's JSON representatoin
        Returns:
            str: the data structure representation
        """
        symbol_json = []
        for i in range(len(self._symbols)):
            if (self._autoupdateviewport):
                self.update_axis_domains(self._symbols[i])
            self._symbols[i].add_all_json(symbol_json, None)

        final_json = {
            "domainX": [self._domainxmin, self._domainxmax],
            "domainY": [self._domainymin, self._domainymax],
            "symbols": symbol_json
        }
        return final_json



from bridges.keypress_listener import *
import traceback

class InputHelper(KeyPressListener):
    """
    @brief This is meant to be an internal class, not something that the library user will use.

    This class provide input device (mouse and keyboard)
    handling for Bridges games
    
    @sa See the Games Tutorials at
    https://bridgesuncc.github.io/tutorials/NonBlockingGame.html
    for more information on keys and mouse device usage.
    
    @author Erik Saule, David Burlinson
    @date 2018, 2019, 2020, 1/2/21
    """

    def __init__(self):
        """
        Constructor. All key initializations done here.
        """
        self.up_key = False
        self.down_key  = False
        self.left_key = False
        self.right_key = False
        self.q_key = False
        self.space_key = False
        self.w_key = False
        self.a_key = False
        self.s_key = False
        self.d_key = False

    def key_press(self, keypress):
        """
        Record a key press event
        Args:
            keypress: keypress event
        """
        try:
            press_type = str(keypress.get("type"))
            key = str(keypress.get("key"))
        except Exception as e:
            traceback.print_tb(e.__traceback__)
            return

        if press_type == "keyup":
            set_to_up = True
        else:
            set_to_up = False

        if press_type == "keydown":
            set_to_down = True
        else:
            set_to_down = False

        if key == "ArrowUp":
            if set_to_up:
                self.up_key = False
            if set_to_down:
                self.up_key = True

        if key == "ArrowDown":
            if set_to_up:
                self.down_key = False
            if set_to_down:
                self.down_key = True

        if key == "ArrowLeft":
            if set_to_up:
                self.left_key = False
            if set_to_down:
                self.left_key = True

        if key == "ArrowRight":
            if set_to_up:
                self.right_key = False
            if set_to_down:
                self.right_key = True

        if key == "q":
            if set_to_up:
                self.q_key = False
            if set_to_down:
                self.q_key = True

        if key == " ":
            if set_to_up:
                self.space_key = False
            if set_to_down:
                self.space_key = True

        if key == "w":
            if set_to_up:
                self.w_key = False
            if set_to_down:
                self.w_key = True

        if key == "a":
            if set_to_up:
                self.a_key = False
            if set_to_down:
                self.a_key = True

        if key == "s":
            if set_to_up:
                self.s_key = False
            if set_to_down:
                self.s_key = True

        if key == "d":
            if set_to_up:
                self.d_key = False
            if set_to_down:
                self.d_key = True

    def status(self):
        print("UP:" + str(self.up()) + " DOWN:" + str(self.down()) + " LEFT:" + str(self.left())
                + " RIGHT:" + str(self.right()))

    def up(self):
        """
        Get the up arrow key.
        Returns:
            the 'Up Arrow' key
        """
        return self.up_key

    def down(self):
        """
        Get the down arrow key.
        Returns:
            the 'Down Arrow' key
        """
        return self.down_key

    def left(self):
        """
        Get the left arrow key.
        Returns:
            the 'Left Arrow' key
        """
        return self.left_key

    def right(self):
        """
        Get the right arrow key.
        Returns:
            the 'Right Arrow' key
        """
        return self.right_key

    def q(self):
        """
        Get the 'q' arrow key.
        Returns:
            the 'q' key
        """
        return self.q_key

    def space(self):
        """
        Get the ' ' (space) key.
        Returns:
            the 'Space' key
        """
        return self.space_key

    def w(self):
        """
        Get the 'w' (space) key.
        Returns:
            the 'w' key
        """
        return self.w_key

    def a(self):
        """
        Get the 'a' (space) key.
        Returns:
            the 'a' key
        """
        return self.a_key

    def s(self):
        """
        Get the 's' (space) key.
        Returns:
            the 's' key
        """
        return self.s_key

    def d(self):
        """
        Get the 'd' (space) key.
        Returns:
            the 'd' key
        """
        return self.d_key








class Shakespeare:
    """
    @brief  A Shakespeare Data source object containing sonnets, poems and plays
   
    This is a convenience class provided for  users who wish to use this
    data source as part of their application. It provides an API that makes
    it easy to access the attributes of this data set.
    
    Refer to tutorial examples to using this data source in data structure
    assignments.

    This object is generally not created by the user, to see how its created check 
    out bridges::data_src_dependent::data_source::get_shakespeare_data()

    For an example, check out https://bridgesuncc.github.io/tutorials/Data_Shakespeare.html

    @author Matthew Mcauaigue, Kalpathi Subramanian

    @date   2018, 12/28/20
    """
    ##
    #    @param title title of sonnet, play or poem
    #    @param type sonnet, play or poem
    #    @param text full text of work
    #
    def __init__(self, title: str = "", type: str = "", text: str = ""):
        self._title = title
        self._type = type
        self._text = text

    ## 
    # @brief Get title
    #
    @property
    def title(self):
        return self._title

    ##
    # @brief set title
    # @param t title of work
    #
    @title.setter
    def title(self, t):
        self._title = t

    ## 
    # @brief Get type (sonnet, play or poem)
    #
    @property
    def type(self):
        return self._type

    ##
    # @brief set type of work (sonnet, play or poem)
    # @param t type of work
    #
    @type.setter
    def type(self, t):
        self._type = t

    ## 
    # @brief Get text of sonnet, play or poem
    #
    @property
    def text(self):
        return self._text

    ##
    # @brief set full text of work
    # @param t full text
    #
    @text.setter
    def text(self, t):
        self._text = t



from bridges.bin_tree_element import *


class BSTElement(BinTreeElement):
    """
    @brief The BSTElement class is the building block for creating binary search trees.
    
    The BSTElement class is the building block for creating binary search tree structures.
    It contains two children (viz., left, right), and a search key, to be used
    in search operations .
    
    BSTElement contains a visualizer (ElementVisualizer) object for setting visual
    attributes (color, shape, opacity, size), necessary for displaying them in a
    web browser.
    
    BST Elements also have a LinkVisualizer object, that is used when they are linked to
    another element, appropriate for setting link attributes, for instance, between
    the current element and its left or right child
    
    @author Kalpathi Subramanian, Mihai Mehedint, Matthew McQuaigue
    
    @date 6/22/16, 1/7/17, 5/17/17, 7/23/19, 2021
    
    @brief This class extends the BinTreeElement class by adding a 'key' value
    for use in a binary search tree implementations.
    
    Binary Search Tree tutorial, https://bridgesuncc.github.io/tutorials/BinarySearchTree.html
    """

    def __init__(self, **kwargs) -> None:
        """
        Constructor bst element
        Kwargs:
            (str) key: The label for the tree element that shows in visualization
            (generic) e: the generic object that the tree element will hold
            (BinTreeElement) left: the tree element assigned to child 0
            (BinTreeElement) right: the tree element assigned to child 1
        Returns:
            None
        """
        if 'e' in kwargs:
            if 'left' in kwargs and 'right' in kwargs:
                super(BSTElement, self).__init__(e=kwargs['e'], left = kwargs['left'], right = kwargs['right'])
            else:
                super(BSTElement, self).__init__(e = kwargs['e'])
        else:
            super(BSTElement, self).__init__()
        if 'key' in kwargs:
            self._key = kwargs['key']

    def get_data_structure_type(self) -> str:
        """
        Get the data structure representation
        Returns:
            str
        """
        return "BinarySearchTree"

    @property
    def key(self) -> str:
        """
        Getter for the bst element key
        Returns:
            str: the key label
        """
        return self._key

    @key.setter
    def key(self, k: str) -> None:
        """
        Setter for the bst element key
        Args:
            k: the key for the element
        Returns:
            None
        """
        self._key = k

    @property
    def left(self):
        """
        Getter for the left element in BST
        Returns:
            BSTElement: the left child of this element
        """
        return super(BSTElement, self).left

    @left.setter
    def left(self, l):
        """
        Setter for the left element in BST
        Args:
            l: the value for the left child to be set as
        """
        self.set_child(0, l)

    @property
    def right(self):
        """
        Getter for the right child in BST
        Returns:
            BSTElement: the right child of the element
        """
        return super(BSTElement, self).right

    @right.setter
    def right(self, r):
        """
        Setter for the right element in BST
        Args:
            r: the value for the right child to be set as
        """
        self.set_child(1, r)

    def get_element_representation(self) -> dict:
        """
        Augment the element with the "key" field
        Returns:
            dict: representing the json of this tree
        """
        orig_json_str = dict(super(BSTElement, self).get_element_representation())
        key_json = {
            'key': self.key
        }
        orig_json_str.update(key_json)
        return orig_json_str



import json
from bridges.symbol import *

class Polyline(Symbol):
    """
    @brief This class defines a polygon and is part of the symbol collection.
    
    A polygon has a set of vertices, with vertices connected by line
    segments. It differs from the polyline in the sense that the last
    and first vertex are connect to close the shape.
    
    @author David Burlinson, Kalpathi Subramanian
    
    @date 2018, 7/23/19, 1/2/21
    
    \sa Shape collection tutorial, https://bridgesuncc.github.io/tutorials/Symbol_Collection.html
    """

    def __init__(self, pts = None):
        """
        Constructor. Sets the number of points in the polyline
        Args:
            pts: number of points to be used in constructing the polyline
        """
        super(Polyline, self).__init__()
        if pts is not None:
            self._points = pts
        else:
            self._points = []

    def get_shape_type(self):
        """
        Get the name of the symbol
        Returns:
            the symbol name
        """
        return "polyline"

    def add_point(self, x, y):
        """
        Add  a point to the polyline
        Args:
            x:  point's x coordinate
            y:  point's y coordinate
        """
        fx = float(x)
        fy = float(y)

        if(fx > float('-inf') and fx < float('inf') and fy > float('-inf') and fy < float('inf')):
            self._points.append(fx)
            self._points.append(fy)

    @property
    def points(self):
        """
        Get the number of points in polyline
        Returns:
            the number of points in symbol.
        """
        return self._points

    @points.setter
    def points(self, pts):
        """
        Set the number of points in polyline
        Args:
            pts: the number of points in symbol.
        """
        self._points = pts


    def get_json_representation(self):

        ds_json = super(Polyline, self).get_json_representation()
        ds_json["points"] = self.points

        return ds_json






class AudioChannel(object):
    """
    @brief This is a class in BRIDGES a channel of audio data
    This class holds a  channel of 8, 16, 24, or 32 bit audio samples.
         
    @author Luke Sloop
        
    @date 2020, 1/31/2020
    """

    def __init__(self, sample_count: int=0, sample_bits: int=32) -> None:
        """
        @brief AudioChannel constructor
        Args:
            sample_count: The total number of samples in this audio channel
            sample_bits:  number of bits per sample (8, 16, 24, 32) 
        """
        
        self.__data = [0 for _ in range(sample_count)]
        self._sample_count = sample_count
        self._sample_bits = 32

    def set_sample(self, index: int, value: int) -> None:
        """
        @brief Set the sample at index. The sample value should be a valid signed integer with get_sample_bits() number of bits.
        Args:
            index:  The index on this channel to set (int)
            value:  The signed integer value to set the sample to (int)
        """
        minmax = (2 ** self.get_sample_bits() // 2) - 1
        if value < -minmax - 1:
            raise ValueError("Sample value out of minimum for signed %d bit integer with value %d" % (self.get_sample_bits(), value))
        if value > minmax:
            raise ValueError("Sample value out of maxmium for signed %d bit integer with value %d" % (self.get_sample_bits(), value))

        self.__data[index] = value

    def get_sample(self, index: int) -> int:
        """
        @brief Get the sample at index. The sample will be a signed integer with get_sample_bits() number of bits.
        Args:
            index: The sample index on this channel to get (int)
        Returns:
            (int) The sample at index that is an integer within the range of an integer of get_sample_bits() number of bits.
        """
        return self.__data[index]

    def get_sample_count(self) -> int:
        """
        @brief Get the number of samples or length of this channel.

        Returns: 
            (int) The number of samples in this channel.
        """
        return self._sample_count

    def get_sample_bits(self) -> int:
        """
        @brief Get the number of bits that make the samples in this channel.

        Will be 8, 16, 24, or 32.
        Returns:
            (int) The number of bits that make the samples in this channel.
        """
        return self._sample_bits



from bridges.element import *
import traceback

class GraphAdjMatrix():
    """
    @brief The GraphAdjMatrix class can be used to represent adjacency matrix based
    graphs in BRIDGES
    
    The GraphAdjMatrix class can be used to represent adjacency matrix based  graphs
    in BRIDGES; it takes 2 generic parameters: (1) K, which is an orderable
    key value used in accessing vertices (in constant time) using a hashmap. This
    permits data sets that need to be accessed by keys that are strings, and
    (2) E, an application defined type, and used in the Edge representation.
    The class is simply a wrapper  around the Java Hashmap class
    and, thus, derives all its operations from it.
    BRIDGES provides methods to visualize the graph  and its contents.
    
    The vertices of the graph are held in a Java hashmap, for near constant time access;
    this lets us use strings or integral ids for vertices. The edges are accessed
    by a second hashmap from each vertex, again assuring near constant access time.
    Each edge contains the terminating vertex id and weight, as defined by  the Edge
    class structure.
    
    Convenience methods are provided to add vertices and edges to the graph. Edges
    are retrieved by using the dual hashmap, given the vertex ids of the edge.
    
    @author Kalpathi Subramanian, Mihai Mehedint
    
    
    \sa Example tutorial at https://bridgesuncc.github.io/tutorials/Graph_AM.html
    
    """

    ##
    #     Constructor
    #
    def __init__(self):
        self._vertices = dict()
        self._matrix = dict()
        self._edge_data = dict()

    ##
    #
    #    This method gets the data structure type
    #
    #    @return  The date structure type as a string
    #
    #
    def get_data_structure_type(self):
        return "GraphAdjacencyMatrix"

    ##
    #     Adds a new vertex to the graph, initializes the  adjacency
    #     list; user is responsible for checking if the vertex already
    #     exists. This method will replace the value for this key
    #
    #     @param k - vertex key value
    #     @param e - user specified data, part of the vertex data
    #
    #
    def add_vertex(self, k, e):
        #  note: it is the user's responsibility to  check
        #  for duplicate vertices
        self.vertices[k] = Element(val = e)
        self.vertices.get(k).label = str(k)

        #  create a hashmap for this vertex
        self._matrix[k] =  dict()
        #  fill up this vertex's row and column elements
        for k, element in self.vertices.items():
            (self._matrix.get(k))[element.label] = 0 #  row
            (self._matrix.get(element.label))[k] =  0 #  col

    ##
    #    Adds a new edge to the graph, adds it to the index corresponding to
    #    the source, destination vertex ids;  this version of the method assumes
    #    an edge weight of 1 (unweighted graph); user is responsible for checking if the
    #    vertices already exist, else an exception is thrown.
    #
    #    @param src - source vertex of edge
    #    @param dest - destination  vertex of edge
    #
    #
    def add_edge(self, src, dest, weight = None):
        #  check to see if the two vertices exist, else
        #  throw an exception
        try:
            if self.vertices.get(src) is None or self.vertices.get(dest) is None:
                raise ValueError("Vertex " + src + " or " + dest + " does not exist! Add the vertex before creating the edge.")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        if weight is not None:
            (self._matrix.get(src))[dest] = weight
        else:
            (self._matrix.get(src))[dest] =  1

    @property
    def vertices(self):
        return self._vertices

    def set_vertex_data(self, src, vertex_data):
        try:
            if self.vertices.get(src) is None:
                raise ValueError("Vertex " + src + " does not exist!")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        self.vertices.get(src).value = vertex_data

    def get_vertex_data(self, src):
        try:
            if self.vertices.get(src) is None:
                raise ValueError("Vertex " + src + " does not exist!")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        return self.vertices.get(src).value

    def set_edge_data(self, src, dest, data):
        try:
            if self.vertices.get(src) is None or self.vertices.get(dest) is None:
                raise ValueError("Vertex " + src + " or " + dest +
					" does not exist! Add the vertex before creating the edge.")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        self._edge_data.get(src)[dest] = data

    def get_edge_data(self, src, dest):
        try:
            if self.vertices.get(src) is None or self.vertices.get(dest) is None:
                raise ValueError("Vertex " + src + " or " + dest +
                                 " does not exist! Add the vertex before creating the edge.")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        return self._edge_data.get(src)[dest]

    ##
    #
    #    Gets the adjacency matrix
    #
    #    @return - the graph's adjacency matrix
    #
    def get_adjacency_matrix(self, key = None):
        if key is None:
            return self._matrix
        else:
            return self._matrix.get(key)

    ##
    #
    #     This is a convenience method to simplify access to the link visualizer;
    #     the method assumes the vertex names point to existing vertices, else an exception
    #     is thrown
    #
    #
    def get_link_visualizer(self, src, dest):
        #  get the source and destination vertex elements
        #  and check to see if they exist
        v1 = self.vertices.get(src)
        v2 = self.vertices.get(dest)
        try:
            if v1 is None or v2 is None:
                raise ValueError("Vertex " + src + " or " + dest + " does not exist! First add the vertices to the graph.")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        return v1.get_link_visualizer(v2)

    ##
    #
    #    This is a convenience method to simplify access to the element visualizer;
    #    the method assumes the vertex name points to an existing vertice, else an
    #    exception is thrown
    #
    #
    def get_visualizer(self, vertex):
        #  get the source and destination vertex elements
        #  and check to see if they exist
        v = self.vertices.get(vertex)
        try:
            if v is None:
                raise ValueError("Vertex " + vertex + " does not exist! First add the vertices to the graph.")
        except Exception as e:
            traceback.print_tb(e.__traceback__)
        return v.visualizer

    ##
    #     Get the JSON representation of the the data structure
    #
    def get_data_structure_representation(self):
        node_map = dict()
        nodes = []
        nodes_json = []
        for key, value in self.vertices.items():
            nodes.append(value)

        for k in range(0, len(nodes)):
            node_map[nodes[k]] = k
            nodes_json.append(nodes[k].get_element_representation())

        links_json = []
        for el_src in self._matrix.items():
            src_vert = self.vertices.get(el_src[0])
            src_indx = node_map.get(src_vert)

            for el_dest in el_src[1].items():
                dest_vert = self.vertices.get(el_dest[0])
                if el_dest[1] > 0:
                    dest_indx = node_map.get(dest_vert)
                    links_json.append(src_vert.get_link_representation(src_vert.get_link_visualizer(dest_vert),
                                                                       str(src_indx), str(dest_indx)))
        json_str = {
            "nodes": nodes_json,
            "links": links_json
        }

        return json_str




class amenityData:
    """
    @brief This class is a helper class to be used with amenities data retrieved from OpenStreet Map data

    The dataset contains amenity id, latitude, longitude, name, and a list of 
    various other values.

    This object is generally not created by the user, to see how its created check 
    out bridges::data_src_dependent::data_source::get_amenity_data()
    
    @sa For an example, check out https://bridgesuncc.github.io/tutorials/Data_Amenity.html
    
    @author Jay Strahler
    
    @date 12/28/20 
    """
    @property
    def id(self): 
        """
        @brief Get id of amenity
        Returns:
            id of amenity
        """
        return self.id_val
    
    @id.setter
    def id(self, id_val):
        """
        @brief Set id of amenity
        Args:
            id_val: id of amenity
        """
        self.id_val = id_val

    @property
    def lat(self): 
        """
        @brief Get latitude of amenity
        Returns:
            latitude of amenity
        """
        return self.lat_val
    
    @lat.setter
    def lat(self, lat_val):
        """
        @brief Set latitude of amenity
        Args:
           lat_val: latitude of amenity to set
        """
        self.lat_val = lat_val

    @property
    def lon(self): 
        """
        @brief Get longitude of amenity
        Returns:
           longitude of amenity
        """
        return self.lon_val
    
    @lon.setter
    def lon(self, lon_val):
        """
        @brief Set longitude of amenity
        Args:
           lon_val: longitude of amenity to set
        """
        self.lon_val = lon_val

    @property
    def name(self): 
        """
        @brief Get name of amenity
        Returns:
            name of amenity
        """
        return self.name_val
    
    @name.setter
    def name(self, name_val):
        """
        @brief Set name of amenity
        Args:
           name_val: name of amenity to set
        """
        self.name_val = name_val

    @property
    def other(self): 
        return self.other_data
    
    @other.setter
    def other(self, other_data):
        """
        @brief Set other of amenity
        Args:
            other_data: other data of amenity
        """
        self.other_data.append(other_data)

    def __init__(self):
        """
        @brief constructor - initialize object
        """
        self.id_val = 0
        self.lat_val = 0
        self.lon_val = 0
        self.name_val = None
        self.other_data = []

class meta:
    """
    @brief This class is a helper class to be used with amenities data retrieved from
    OpenStreet Map data

    This meta dataset contains the base information of the query requested.
    This includes minimum latitude, minimum longitude, maximum latitude, 
    maximum longitude, and a count of the total values returned.
    
    @author Jay Strahler
    
    @date 12/28/20 
    """
    @property
    def minLat(self): 
        return self.minLat_val
    
    @minLat.setter
    def minLat(self, minLat_val):
        self.minLat_val = minLat_val

    @property
    def minLon(self): 
        return self.minLon_val
    
    @minLon.setter
    def minLon(self, minLon_val):
        self.minLon_val = minLon_val

    @property
    def maxLat(self): 
        return self.maxLat_val
    
    @maxLat.setter
    def maxLat(self, maxLat_val):
        self.maxLat_val = maxLat_val

    @property
    def maxLon(self): 
        return self.maxLon_val
    
    @minLon.setter
    def maxLon(self, maxLon_val):
        self.maxLon_val = maxLon_val


    @property
    def count(self):
        return self.count_val
    
    @count.setter
    def count(self, count_val):
        self.count_val = count_val

    def __init__(self, minLat_val, minLon_val, maxLat_val, maxLon_val, count_val):
        self.minLat_val = minLat_val
        self.minLon_val = minLon_val
        self.maxLat_val = maxLat_val
        self.maxLon_val = maxLon_val
        self.count_val = count_val

class amenities:
    """
    @brief This class is a helper class to be used with amenities data retrieved from
    OpenStreet Map data

    The dataset contains a list of amenityData objects, and the meta object. 

    For an example, check out https://bridgesuncc.github.io/tutorials/Data_Amenity.html
    
    @author Jay Strahler
    
    @date 12/28/20 
    """
    @property
    def data(self):
        return self.data_val

    @data.setter
    def data(self, data_val):
        self.data_val.append(data_val)

    @property
    def meta(self):
        return self.meta_val

    @meta.setter
    def meta(self, meta_val):
        self.meta_val = meta_val

    def __init__ (self):
        self.data_val = []
        self.meta_val = None





class USCities():

    def __init__(self, city:str = None, state:str = None, country:str = None, lat:float = None, lon:float = None, elevation:int = None, population:int = None, timezone:str = None):
        self._city = city
        self._state = state
        self._country = country
        self._lat = lat
        self._lon = lon
        self._elevation = elevation
        self._population = population
        self._timezone = timezone


    @property
    def city(self):
        return self._city

    @city.setter
    def city(self, city):
        self._city = city

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, state):
        self._state = state

    @property
    def country(self):
        return self._city

    @country.setter
    def country(self, country):
        self._country = country

    @property
    def lat(self):
        return self._lat

    @lat.setter
    def lat(self, lat):
        self._lat = lat

    @property
    def lon(self):
        return self._lon

    @lon.setter
    def lon(self, lon):
        self._lon = lon

    @property
    def elevation(self):
        return self._elevation

    @elevation.setter
    def elevation(self, elevation):
        self._elevation = elevation

    @property
    def population(self):
        return self._population

    @population.setter
    def population(self, population):
        self._population = population

    @property
    def timezone(self):
        return self._timezone

    @timezone.setter
    def timezone(self, timeone):
        self._timezone = timezone







import json
from bridges.symbol import *


class Circle(Symbol):
    """
    @brief This class defines a circle and is part of the symbol collection.
    A circle  has a radius and a center, which is also its location
    
    @author Matthew Mcquaigue
    @date 2018, 7/23/19
    
    Shape collection tutorial, https://bridgesuncc.github.io/tutorials/Symbol_Collection.html
    """

    def __init__(self, **kwargs) -> None:
        """
        Constructor for a circle symbol
        Args:
            (int) locx: x location of the circle
            (int) locy: y location of the circle
            (float) r: radius of the circle
        Returns:
            None
        Raises:
            ValueError: if the radius is not a positive number
        """
        super(Circle, self).__init__()
        if 'locx' in kwargs and 'locy' in kwargs:
            self.locx = kwargs['locx']
            self.locy = kwargs['locy']
        else:
            self.locx = 0.0
            self.locy = 0.0
        if 'r' in kwargs:
            if kwargs['r'] < 0:
                raise ValueError("Radius value needs to be positive")
            self.radius = kwargs['r']
        else:
            self.radius = 10

    def get_shape_type(self) -> str:
        """
        Gets the name of the shape/symbol
        Returns:
            str: representing the name
        """
        return "circle"

    @property
    def radius(self) -> float:
        """
        Getter function for the radius of this circle
        Returns:
            float: representing radius
        """
        return self._radius

    @radius.setter
    def radius(self, r) -> None:
        """
        Setter for the radius of this circle
        Args:
            (float) r: radius to be set
        Returns:
            None
        Raises:
            ValueError: if the radius is not a positive number
        """
        if r < 0:
            raise ValueError("Illegal value for radius. Must be positive")
        self._radius = r

    def set_circle(self, locx, locy, r) -> None:
        """
        Set the location and size of the circle
        Args:
            (int) locx: x coordinate of the center of the circle
            (int) locy: y coordinate of the center of the circle
            (float) r: radius of the circle
        Returns:
            None
        Raises:
            ValueError: if the radius is not a positive number
        """
        self.locx = locx
        self.locy = locy
        if r < 0:
            raise ValueError("Radius value needs to be positive")
        self.radius = r

    def get_json_representation(self) -> dict:
        """
        Get the json representation of the Circle object
        Returns:
            dict: representing the JSON 
        """
        ds_json = super(Circle, self).get_json_representation()
        ds_json["r"] = self.radius
        ds_json['center'] = [self.locx, self.locy]

        return ds_json



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="bridges",
    version="3.3.1",
    author="BRIDGES",
    author_email="bridges.uncc@gmail.com",
    description="Bridging Real-world Infrastructure Designed to Goal-align, Engage, and Stimulate, an NSF TUES",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/BridgesUNCC/bridges-python",
    packages=setuptools.find_packages(exclude=("tests")),
    python_requires='>3.5.2',
    install_requires=[
        'requests>=2.21.0',
        'webcolors>=1.8.1',
        'python-socketio[client]>=4.3.0',
        'SPARQLWrapper>=1.8.4',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



#!/usr/bin/env python
from bridges.element import *

class Array():
    """
    @brief This class can be used to create arrays of type Element<E>.
    
    @author 	Matthew McQuaigue
    
    @date  	10/8/16, 6/09/19
    
    This class can be used to create arrays of type Element<E>  where E
    is a generic object representing application specific data.
    
    Arrays are internally represented as 1D arrays; currently 1D, 2D  and
    3D arrays are supported.
    
    Example Tutorial at: https://bridgesuncc.github.io/tutorials/Array.html (1D, 2D, and 3D Array)
    """

    dims = [1,1,1] #used for setting size of array based on dimensions
    def __init__(self, **kwargs):
        """
        Array constructor
        Kwargs:
            dims: size of each dimension (array)
            num_dims: The dimensions of the array (1-3). Defaults to 1 dimension (int)
        Returns:
            None
        """
        self._dims = [1, 1, 1]
        self._array_data = []
        self._num_dims = 1
        self._size = 0
        if 'num_dims' in kwargs:
            if 'dims' in kwargs:
                self.set_size(kwargs['num_dims'], kwargs['dims'])

    @property
    def num_dims(self) -> int:
        """
        Getter for representing the number of dimensions in the array
        Returns:
            int: number of dimensions
        """
        return self._num_dims

    @num_dims.setter
    def num_dims(self, value: int) -> None:
        """
        Setter function for the number of dimensions for the array
        Args:
            value: An integer for the number of dimensions (Between 1 and 3 inclusive)
        Returns:
            None
        Raises:
            ValueError: if dimension passed in is < 1 or > 3
        """
        if value > 3:
            raise ValueError("Invalid number of dimensions. Only 1D, 2D and 3D arrays supported at this time")
        self._num_dims = value

    @property
    def size(self) -> int:
        """
        Getter for representing the size of array
        Returns:
            int: the size
        """
        return self._size

    @size.setter
    def size(self, sz: int) -> None:
        """
        Setter for representing the size of the array
        Args:
            (int) sz: The size to be set for array
        Returns:
            None
        """
        self._size = sz

    def set_size(self, nd: int, dim: list) -> None:
        """
        Set the size of each dimensions; also allocates  array space
        Args:
            (int) nd: number of dimensions
            (list) dim: size of each dimension
        Returns:
            None
        """
        if dim[0] <=0 or dim[1] <=0 or dim[2] <= 0:
            raise ValueError("Invalid dimension value, must be  positive")

        self._dims[0] = dim[0]
        self._dims[1] = dim[1]
        self._dims[2] = dim[2]
        self.num_dims = nd
        self.size = dim[0]*dim[1]*dim[2]

        for k in range(0, self.size):
            self._array_data.append(Element())


    def get_data_structure_type(self) -> str:
        """
        Gets the data structure type
        Raises:
            ValueError: if number of dimensions is < 1 or > 3
        Returns:
            str: type of data structure
        """
        if (self.num_dims >= 1) and (self.num_dims <= 3):
            return "Array"
        else:
            raise ValueError("Invalid number of dimensions. Only 1D, 2D and 3D arrays supported at this time")

    def get_dimensions(self) -> list:
        """
        Get the size of each dimensions
        Returns:
            list: size of each dimension
        """
        return self._dims

    def get_element(self, *args, **kwargs):
        """
        Getter function for an element in the array at given position
        args:
            (int) x,y,z,: indices
        Kwargs:
            (int) index: the index of array to get in array
            (int) x: column index into array
            (int) y: row index into array
            (int) z: slice index into array
        Returns:
            Element: the element at position given
        """
        if len(args) != 0:
            if len(args) != self._num_dims:
                raise IndexError("args must be same length as dims")
            if len(args) == 1:
                return self._array_data[args[0]]
            if len(args) == 2:
                return self._array_data[args[0] * self._dims[0] + args[1]]
            if len(args) == 3:
                return self._array_data[args[0] * self._dims[0] * self._dims[1] + args[1] * self._dims[0] + args[2]]

        elif 'index' in kwargs:
            return self._array_data[kwargs['index']]
        elif 'x' in kwargs and 'y' in kwargs:
            if 'z' in kwargs:
                return self._array_data[kwargs['z'] * self._dims[0] *
                                        self._dims[1] + kwargs['y'] * self._dims[0] + kwargs['x']]
            else:
                return self._array_data[kwargs['y'] * self._dims[0] + kwargs['x']]

    def set_element(self, *args, **kwargs):
        """
        Setter function for an element in the array at given position
        Args:
            (int) x,y,z: indices
            (Element) el: element object to be assigned to index, always last position arg if using unnamed args
        Kwargs:
            (Element) el: element object to be assigned to index
            (int) index: the index of array to get in array
            (int) x: column index into array
            (int) y: row index into array
            (int) z: slice index into array
        Returns:
            None
        """
        if len(args) == 0 and kwargs["element"] is None:
            raise RuntimeError("Must have at least 1 unnamed element parameter or named element parameter")
        if len(args) != self._num_dims + 1 and len(args) != 1:
            raise RuntimeError("Must used proper indices if not using kwargs")
        else:
            if len(args) == 2:
                self._array_data[args[0]] = args[1]
            if len(args) == 3:
                self._array_data[args[1] * self._dims[0] + args[0]] = args[2]
            if len(args) == 4:
                self._array_data[args[2] * self._dims[0] * self._dims[1] + args[1] * self._dims[0] + args[0]] = args[3]

        if 'index' in kwargs:
            self._array_data[kwargs['index']] = args[0]
        if 'x' in kwargs and 'y' in kwargs:
            if 'z' in kwargs:
                self._array_data[kwargs['z']*self._dims[0]*self._dims[1] + kwargs['y']*self._dims[0] + kwargs['x']] = args[0]
            else:
                self._array_data[kwargs['y']*self._dims[1]+ kwargs['x']] = args[0]

    def __getitem__(self, item):
        """
        :param  (tuple) item: indices
        :return: element at index
        """
        if type(item) == int:
            if self._num_dims != 1:
                raise IndexError("invalid index, pass a tuple of same length of number of dimensions of your array")
            return self.get_element(item)
        if len(item) != self._num_dims:
            raise IndexError("invalid index, pass a tuple of same length of number of dimensions of your array")

        return self.get_element(*item)

    def __setitem__(self, key, value):
        if type(key) == int:
            if self._num_dims != 1:
                raise IndexError("invalid index, pass a tuple of same length of number of dimensions of your array")
            self.set_element(key, value)
            return
        if len(key) != self._num_dims:
            raise IndexError("invalid index, pass a tuple of same length of number of dimensions of your array")
        self.set_element(*key, value)

    def get_data_structure_representation(self) -> dict:
        """
        Generating the JSON string for a bridges array object
        Returns:
            dict: the dict that will represent the json when dumped
        """
        #add json representation for each element to dict
        i = 0
        nodes_json = []
        while i < self.size:
            if self._array_data[i] is not None:
                nodes_json.append(self._array_data[i].get_element_representation())
            i += 1
        json_dict = {
            "nodes": nodes_json,
            "dims": [self._dims[0], self._dims[1], self._dims[2]]
        }
        return json_dict
